#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import uuid

class ColonyException(Exception):
    """
    The top level colony exception, this is the main exception
    for the complete colony infra-structure and all the other
    exceptions should inherit from this one.
    """

    message = None
    """ The exception's message, to be used latter for runtime
    based diagnostics """

    code = None
    """ The internal error code to be used in objective
    serialization of the error (eg: HTTP) """

    headers = None
    """ Optional list of MIME compliant headers to be sent
    in a client/server paradigm """

    meta = None
    """ The meta information associated with the error
    that should be considered private in context, this
    should be structured data ready to be serializable """

    def __init__(self, message = None, **kwargs):
        Exception.__init__(self)
        self.message = message
        self.code = kwargs.get("code", 500)
        self.headers = kwargs.get("headers", None)
        self.meta = kwargs.get("meta", None)
        self._uid = None

    def __str__(self):
        """
        Returns the string representation of the class.

        :rtype: String
        :return: The string representation of the class.
        """

        if self.message == None: return Exception.__str__(self)
        return "Colony exception - %s" % self.message

    def __unicode__(self):
        """
        Returns the unicode representation of the class.

        :rtype: String
        :return: The unicode representation of the class.
        """

        return self.__str__()

    @property
    def uid(self):
        """
        The unique identifier of the current exception may
        be used safely from a global/universal point of view.

        :rtype: String
        :return: The global unique identifier of the current
        exception entity.
        """

        if self._uid: return self._uid
        self._uid = uuid.uuid4()
        return self._uid

class OperationalError(ColonyException):
    """
    Error raised for a runtime error and as a result
    of an operational routine that failed.
    This should not be used for coherent development
    bugs, that are raised continuously.
    """

    pass

class AssertionError(OperationalError):
    """
    Error raised for failure to meet any pre-condition or
    assertion for a certain data set.
    """

    def __init__(self, *args, **kwargs):
        kwargs["message"] = kwargs.get("message", "Assertion of data failed")
        kwargs["code"] = kwargs.get("code", None)
        OperationalError.__init__(self, *args, **kwargs)

class PluginSystemException(ColonyException):
    """
    The abstract plugin system exception class, that
    should represent all the exception that are associated
    with the plugin system environment.
    """

    def __init__(self, message):
        """
        Constructor of the class.

        :type message: String
        :param message: The message to be printed.
        """

        ColonyException.__init__(self, message = message)

    def __str__(self):
        """
        Returns the string representation of the class.

        :rtype: String
        :return: The string representation of the class.
        """

        return "Plugin system exception - %s" % self.message

class PluginClassNotAvailable(PluginSystemException):
    """
    The plugin class not available class.
    """

    def __str__(self):
        """
        Returns the string representation of the class.

        :rtype: String
        :return: The string representation of the class.
        """

        return "Plugin class not available - %s" % self.message

class InvalidCommand(PluginSystemException):
    """
    The invalid command class.
    """

    def __str__(self):
        """
        Returns the string representation of the class.

        :rtype: String
        :return: The string representation of the class.
        """

        return "Invalid command - %s" % self.message

class InvalidArgument(PluginSystemException):
    """
    The invalid argument class, meaning that some passed
    parameter/argument is not valid under the context of
    the current execution.
    """

    def __str__(self):
        """
        Returns the string representation of the class.

        :rtype: String
        :return: The string representation of the class.
        """

        return "Invalid argument - %s" % self.message

class SecurityError(PluginSystemException):
    """
    The security (kind) error, that represents an issue
    related with something that may be able to compromise
    the security of the system.
    """

    def __str__(self):
        """
        Returns the string representation of the class.

        :rtype: String
        :return: The string representation of the class.
        """

        return "Security error - %s" % self.message

class OperationNotComplete(PluginSystemException):
    """
    The operation not complete class, meaning that the
    some operation has been blocked in the middle of execution.
    """

    def __str__(self):
        """
        Returns the string representation of the class.

        :rtype: String
        :return: The string representation of the class.
        """

        return "Operation not complete - %s" % self.message

class OperationRestart(PluginSystemException):
    """
    The operation restart class, meaning that the current
    executing operation should be restart by its owner (up
    in the call stack).
    """

    delay = 0
    """ The delay (in seconds) to be used in the waiting
    before the restart of the operation """

    exception = None
    """ The original exception that has triggered the need
    for an operation restart """

    def __init__(self, message, delay = 0, exception = None, **kwargs):
        """
        Constructor of the class.

        :type message: String
        :param message: The message to be printed.
        :type delay: float
        :param delay: The delay (in seconds) to be used in
        the waiting before the restart of the operation.
        :type exception: Exception
        :param exception: The original exception that has triggered
        the need for an operation restart.
        """

        PluginSystemException.__init__(self, message, **kwargs)
        self.delay = delay
        self.exception = exception

    def __str__(self):
        """
        Returns the string representation of the class.

        :rtype: String
        :return: The string representation of the class.
        """

        delay_s = " (%0.1f seconds delay)" % self.delay if self.delay else ""
        return "Operation restart%s - %s" % (delay_s, self.message)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from . import aes_util
from . import bank_util
from . import barcode_util
from . import cache_util
from . import call_util
from . import control_util
from . import country_util
from . import crypt_util
from . import encode_util
from . import file_util
from . import gtin_util
from . import host_util
from . import import_util
from . import lazy_util
from . import list_util
from . import logging_util
from . import map_util
from . import math_util
from . import number_util
from . import object_util
from . import observer_util
from . import os_util
from . import path_util
from . import protection_util
from . import quote_util
from . import round_util
from . import scheduling_util
from . import size_util
from . import stack_util
from . import string_buffer_util
from . import string_util
from . import structures_util
from . import test_util
from . import time_util
from . import update_thread_util
from . import verify_util
from . import version_util
from . import visitor_util

from .aes_util import AesCipher
from .barcode_util import encode_2_of_5, encode_code_128, encode_code_39
from .cache_util import DataCacheMap
from .call_util import execute_retries, call_safe
from .control_util import calculate_tax_number_control_value, calculate_id_number_control_value
from .country_util import COUNTRIES, country_get
from .crypt_util import password_crypt, password_match, password_strength, md5_crypt,\
    generate_hash_digest_map
from .encode_util import encode_two_complement_string, decode_two_complement_string
from .file_util import FileRotator, FileContext, TransactionContext, FileImmediateContext,\
    FileTransactionContext
from .host_util import get_hostname, get_hostname_local, get_address_ip4, get_address_ip4_force,\
    get_address_ip4_all, get_address_ip6, get_address_ip6_force, get_address_ip6_all,\
    get_addresses_ip4, get_addresses_ip6, get_addresses_family, get_address_tuples,\
    ip4_address_from_network, ip4_address_to_network, ip6_address_from_network,\
    ip6_address_to_network
from .import_util import reload_import
from .lazy_util import LazyClass, LazyIteratorClass, is_lazy, Lazy, LazyIterator
from .list_util import list_intersect, list_extend, list_no_duplicates
from .logging_util import getLogger, getLevelName, DummyLogger, StreamHandler, Formatter
from .map_util import map_clean, map_get, map_copy, map_copy_deep, map_duplicate,\
    map_remove, map_extend, map_flatten, map_check_parameters, map_get_value_cast,\
    map_get_values, map_output, map_normalize
from .math_util import ceil_integer, greatest_common_divisor, fast_exponentiation,\
    item_set_total, item_set_percentage
from .number_util import get_number_length, get_digit, to_fixed
from .object_util import object_attribute_names, object_attribute_values, object_flatten,\
    object_print_list, object_print
from .observer_util import unique, notify, message, action, progress, register_g,\
    unregister_g, notify_g
from .os_util import kill_process
from .path_util import SEPARATOR, normalize_path, align_path, copy_directory, copy_link, copy_file,\
    remove_directory, link, link_copy, ensure_file_path, is_parent_path, relative_path
from .protection_util import public, Protected
from .quote_util import quote, quote_plus, unquote, unquote_plus, url_encode
from .round_util import roundi, rounds, roundt, round_apply, round_unapply, round_is_new
from .scheduling_util import SCHEDULING_MAX, Scheduler
from .size_util import size_round_unit
from .stack_util import get_instance_module_directory, get_call_module_directory
from .string_buffer_util import StringBuffer
from .string_util import xor_string_value, to_underscore, to_camelcase, pluralize,\
    capitalize_all, join
from .structures_util import Decimal, JournaledList, OrderedMap, OrderedMapIterator,\
    MultipleValueMap, FormatTuple, FileReference, is_dictionary
from .test_util import ColonyTestCase
from .time_util import SIMPLE_VALUE, BASIC_VALUE, EXTENDED_VALUE, EXTENDED_SIMPLE_VALUE,\
    MINIMIZE_MULTIPLE, MINIMIZE_UNIQUE, format_seconds_smart, format_seconds, timestamp_datetime
from .update_thread_util import UpdateThread
from .verify_util import verify, verify_equal, verify_not_equal, verify_type, verify_many
from .version_util import version_cmp, version_is_concrete
from .visitor_util import visit, dispatch_visit

from .bank_util import calculate_control_value as calculate_control_value_bank
from .gtin_util import calculate_control_value as calculate_control_value_gtin

from .import_util import __import__



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "Luís Martinho <lmartinho@hive.pt> & João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from colony.base import legacy

from . import string_buffer_util

DIGIT_ENCODING_MAP = {
    "0" : "NNWWN",
    "1" : "WNNNW",
    "2" : "NWNNW",
    "3" : "WWNNN",
    "4" : "NNWNW",
    "5" : "WNWNN",
    "6" : "NWWNN",
    "7" : "NNNWW",
    "8" : "WNNWN",
    "9" : "NWNWN"
}
""" The digit encoding map converting the digit into
N(arrow) and W(ide) combinations, useful for 2 of 5 encoding  """

START_CODE_2_OF_5 = "NnNn"
""" The start code for 2 of 5 """

END_CODE_2_OF_5 = "WnN"
""" The end code for 2 of 5 """

START_CODES_CODE_128 = {
    "A" : 103,
    "B" : 104,
    "C" : 105
}
""" The code 128 start codes for each code set """

END_CODE_CODE_128 = 106
""" The end code for code 128 """

def encode_2_of_5(string_value):
    """
    Encodes the provided string value into the 2 of 5
    barcode encoding to be used with the most prominent
    true type fonts for display.

    :type string_value: String
    :param string_value: The string value to be encoded into
    the 2 of 5 barcode string representation.
    :rtype: String
    :return: The generated 2 of 5 barcode string representation.
    :see: http://en.wikipedia.org/wiki/Interleaved_2_of_5
    """

    # starts the encoded buffer with the default
    # start code for the 2 of 5
    encoded_buffer = string_buffer_util.StringBuffer()
    encoded_buffer.write(START_CODE_2_OF_5)

    # retrieves the length of the string value and the
    # uses it to check if the string value length is
    # of type odd or even
    string_value_length = len(string_value)
    is_even_length = string_value_length % 2 == 0

    # in case the string value does not contains an even
    # number of digits a zero character is prepended to
    # the string value (so that it's possible to codify it)
    if not is_even_length: string_value = "0" + string_value

    # resets the index counter
    index = 0

    # iterates over the range of indexes of the
    # string value range
    while index < string_value_length:
        # retrieves the first and second digits
        # for the current string value index
        first_digit = string_value[index]
        second_digit = string_value[index + 1]

        # encodes the first and second digits
        # using the the digit encoding map
        first_digit_encoded = DIGIT_ENCODING_MAP.get(first_digit, None)
        second_digit_encoded = DIGIT_ENCODING_MAP.get(second_digit, None)

        # in case at least one of the digits is not valid
        # (not possible to encoded) it, this is an error
        # situation and a runtime error must be raised
        if first_digit_encoded == None or second_digit_encoded == None: raise RuntimeError("Value is not serializable in 2 of 5 encoding")

        # interleaves both the first and second encoded
        # digits and then writes the result to the encoded buffer
        interleaved_digits_string = _interleave_digits(first_digit_encoded, second_digit_encoded)
        encoded_buffer.write(interleaved_digits_string)

        # increments the index counter in two values
        # so that so that the s
        index += 2

    # writes the end of the 2 of 5 encoding
    # value to the encoded buffer and then
    # retrieves the (final) encoded value
    encoded_buffer.write(END_CODE_2_OF_5)
    encoded_value = encoded_buffer.get_value()

    # returns the 2 of 5 encoded value
    return encoded_value

def encode_code_128(value, code_set = "A"):
    """
    Encodes the provided string value into the code 128
    barcode encoding to be used with the most prominent
    true type fonts for display.

    :type string_value: String
    :param string_value: The string value to be encoded into
    the code 128 barcode string representation.
    :rtype: String
    :return: The generated code 128 barcode string representation.
    :see: http://en.wikipedia.org/wiki/Code_128
    """

    # encodes the value into the proper representation
    # using the requested code set, in case no valid code
    # set is provided raises a runtime error
    if code_set == "A": character_values = _encode_code_set_a(value)
    elif code_set == "B": character_values = _encode_code_set_b(value)
    elif code_set == "C": character_values = _encode_code_set_c(value)
    else: raise RuntimeError("Specified code set not supported %s" % code_set)

    # computes and appends the check digit
    # using the current character values list
    check_digit = _calculate_check_digit(character_values)
    character_values.append(check_digit)

    # appends the final end code, to finish
    # the set of character values
    character_values.append(END_CODE_CODE_128)

    # retrieves the character string for the
    # code 128 values and returns it as the
    # final encoded unicode string
    character_string = _get_character_string(character_values)
    return character_string

def encode_code_39(value):
    """
    Encodes the provided string value into the code 39
    barcode encoding to be used with the most prominent
    true type fonts for display.

    :type string_value: String
    :param string_value: The string value to be encoded into
    the code 39 barcode string representation.
    :rtype: String
    :return: The generated code 39 barcode string representation.
    :see: http://en.wikipedia.org/wiki/Code_39
    """

    return "*" + value + "*"

def _interleave_digits(first_digit, second_digit):
    """
    Interleaves both 2 of 5 encoded digits by creating
    a composite string that represents the interleaved
    version of them.

    This process is specific for the general version of
    the 2 of 5 true type font encoding.

    :type first_digit: String
    :param first_digit: The first digit encoded in the
    2 of 5 normal encoding.
    :type second_digit: String
    :param second_digit: The second digit encoded in the
    2 of 5 normal encoding.
    :rtype: String
    :return: The interleaved representation of the concatenation
    of both digits in the normal 2 of 5 barcode format.
    """

    # creates the list that will hold the
    # various interleaved letters
    interleaved_leetters = []

    # iterates over the default range
    # of a digit encoding in 2 of 5
    for index in range(5):
        # retrieves the first and the second
        # letters from the first and second digits
        first_letter = first_digit[index]
        second_letter = second_digit[index]

        # converts the second letter to lowercase ands
        # then creates the interleaved letters sequence
        # by appending the first letter to the lowercase
        # version of the second letter and then appends
        # it to the interleaved letter list
        second_letter_lower = second_letter.lower()
        interleaved_letters = first_letter + second_letter_lower
        interleaved_leetters.append(interleaved_letters)

    # joins the set of interleaved letters to obtain the
    # "final" interleaved digits string, then returns it
    interleaved_digits_string = "".join(interleaved_leetters)
    return interleaved_digits_string

def _calculate_check_digit(character_values):
    """
    Computes the check digit based on a modulus 103 checksum,
    the algorithm is quite simple, the idea is to weight the
    value of each position over its own index position.

    :type character_values: List
    :param character_values: The list of character values to
    calculate the check digit (checksum).
    :rtype: int
    :return: The resulting check digit for the provided set
    of character values.
    """

    # starts the checksum value with the first
    # value and then retrieves the string
    # containing the remaining values
    checksum = character_values[0]
    remaining_values = character_values[1:]
    remaining_values_length = len(remaining_values)

    # iterates over the range of the remaining
    # value length to calculate the checksum
    for index in range(remaining_values_length):
        # retrieves the current value and multiplies
        # it over it's own index value plus one, that
        # will be its weighted value (value to be added
        # to the checksum integer)
        value = remaining_values[index]
        weighted_value = value * (index + 1)
        checksum += weighted_value

    # runs the modulus operation on the checksum
    # and returns it as its check digit
    check_digit = checksum % 103
    return check_digit

def _encode_code_set_a(string_value):
    """
    Encodes the provided string value into a list of code 128
    values using the code set a.

    :type string_value: String
    :param string_value: The string value to be converted into the
    list of code 128 values.
    :rtype: List
    :return: The list containing a set of code 128 numeric
    values representing the requested string in code set a.
    """

    # retrieves the start code for the current
    # encoding and starts the character values
    # list with the start code
    start_code = START_CODES_CODE_128["A"]
    character_values = [start_code]

    # iterates over all the characters contained
    # in the string value to encode them
    for character in string_value:
        # retrieves the ordinal value of the character
        # and checks if it's a valid one
        character_ascii_value = ord(character)
        if character_ascii_value > 95: raise RuntimeError("Code set does not support character '%s' " % character)

        # "calculates" the appropriate code 128 code set a
        # for the value and appends it to the character values
        character_code_128_value = character_ascii_value > 32 and character_ascii_value - 32 or character_ascii_value + 64
        character_values.append(character_code_128_value)

    # returns the list containing the code set
    # a encoded values
    return character_values

def _encode_code_set_b(string_value):
    """
    Encodes the provided string value into a list of code 128
    values using the code set b.

    :type string_value: String
    :param string_value: The string value to be converted into the
    list of code 128 values.
    :rtype: List
    :return: The list containing a set of code 128 numeric
    values representing the requested string in code set b.
    """

    # retrieves the start code for the current
    # encoding and starts the character values
    # list with the start code
    start_code = START_CODES_CODE_128["B"]
    character_values = [start_code]

    # iterates over all the characters contained
    # in the string value to encode them
    for character in string_value:
        # retrieves the ordinal value of the character
        # and checks if it's a valid one
        character_ascii_value = ord(character)
        if character_ascii_value < 32: raise RuntimeError("Code set does not support character '%s' " % character)

        # "calculates" the appropriate code 128 code set b
        # for the value and appends it to the character values
        character_code_128_value = character_ascii_value - 32
        character_values.append(character_code_128_value)

    # returns the list containing the code set
    # b encoded values
    return character_values

def _encode_code_set_c(string_value):
    """
    Encodes the provided string value into a list of code 128
    values using the code set c.

    :type string_value: String
    :param string_value: The string value to be converted into the
    list of code 128 values.
    :rtype: List
    :return: The list containing a set of code 128 numeric
    values representing the requested string in code set c.
    """

    # retrieves the start code for the current
    # encoding and starts the character values
    # list with the start code
    start_code = START_CODES_CODE_128["C"]
    character_values = [start_code]

    # resets the index value
    index = 0

    # iterates over the range of string value
    # to access all of its characters
    while index < len(string_value) - 1:
        # retrieves the current character slice and
        # parses it as an integer to append it to
        # the character values list
        characters_slice = string_value[index:index + 2]
        character_value = int(characters_slice)
        character_values.append(character_value)
        index += 2

    # returns the list containing the code set
    # c encoded values
    return character_values

def _get_character_string(character_values):
    """
    Converts the provided list of code 128 characters
    in an integer format into the appropriate unicode
    representation of them in the code 128 true type
    font format.

    :type character_values: List
    :param character_values: The list of code 128 integer
    values to be converted into the unicode string.
    :rtype: String
    :return: The converted unicode string, ready to be
    used in the true type font.
    """

    # creates the buffer to hold the various partial
    # encoded value that will constitute the "final"
    # character string
    character_buffer = string_buffer_util.StringBuffer()

    # iterates over all the characters in the list of
    # code 128 character values
    for character_value in character_values:
        # converts the value offset according to the position
        # in the conversion table
        if character_value < 95: ordinal_value = character_value + 32
        else: ordinal_value = character_value + 100

        # converts the ordinal value into an
        # unicode encoded value and adds it
        # to the character buffer
        character = legacy.unichr(ordinal_value)
        character_buffer.write(character)

    # retrieves the character string as the value from the
    # character buffer and then returns the value
    character_string = character_buffer.get_value()
    return character_string



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import copy

def list_intersect(first_list, second_list):
    """
    Intersects the given lists, returning a list with
    the elements contained in both lists.

    :type first_list: List
    :param first_list: The first list to be used in intersection.
    :type second_list: List
    :param second_list: The second list to be used in intersection.
    :rtype: List
    :return: The list containing the elements contained
    in both lists (intersection).
    """

    # returns the intersection resulting list
    return [value for value in first_list if value in second_list]

def list_extend(base_list, extension_list, copy_base_list = True):
    """
    Extends the list with the the extension list,
    returning a list resulting of the merge of both list.

    Duplicates are avoided to remove additional elements.
    The base list may be changed or left untouched based on
    the copy base list flag.

    :type base_list: List
    :param base_list: The list to be used as base for
    the merge.
    :type extension_list: List
    :param extension_list: The list to be used to extend the base
    one.
    :type copy_base_list: bool
    :param copy_base_list: If the base list should be copied before
    being extended in order to avoid loss of data.
    :rtype: List
    :return: The list that result of the merge of both lists.
    """

    # copies the base list to create the initial result list (optional)
    result_list = copy_base_list and copy.copy(base_list) or base_list

    # creates the list of values that are "new" to the base
    # list (in order to avoid duplicates)
    filtered_list = [value for value in extension_list if not value in base_list]

    # extends the result list with the filtered list (new elements)
    result_list.extend(filtered_list)

    # returns the result list
    return result_list

def list_exclude(base_list, exclusion_list, copy_base_list = True):
    """
    Excludes a series of items from the given base list,
    the resulting list (without the items) is returned.
    The base list may be changed or left untouched based on
    the copy base list flag.

    :type base_list: List
    :param base_list: The list to be used as base for
    the exclusion.
    :type exclusion_list: List
    :param exclusion_list: The list to be used as model for
    the exclusion of items
    :type copy_base_list: bool
    :param copy_base_list: If the base list should be copied before
    the exclusion process in order to avoid loss of data.
    :rtype: List
    :return: The list that result of the exclusion of the given
    list from the provided base list.
    """

    # copies the base list to create the initial result list (optional)
    result_list = copy_base_list and copy.copy(base_list) or base_list

    # iterates over all the items in the exclusion list
    # to remove them from the "result" list
    for exclusion_item in exclusion_list:
        # removes the exclusion item from  the result list
        result_list.remove(exclusion_item)

    # returns the result list
    return result_list

def list_no_duplicates(list):
    """
    Removes the duplicated values from the given list.
    This method is expensive and should be used carefully.

    :type list: List
    :param list: The list to heave it's duplicates removed.
    :rtype: List
    :return: The list with the duplicates removed.
    """

    # creates the initial result list to hold the results
    result_list = []

    # iterates over all the values in the list
    for value in list:
        # in case the value already exists
        # in the result list (duplicate)
        if value in result_list:
            # continues the loop
            continue

        # adds the value to the result list
        result_list.append(value)

    # returns the result list
    return result_list



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import unittest

class Test(object):
    """
    The base and abstract test class from which all the
    colony test back end class will inherit from. It
    should contain the basic infra-structure for a rapid
    creation of tests to be executed in colony.
    """

    plugin = None
    """ The reference to the plugin that "owns" this
    test object, this may be used to reference the
    top level manager functions """

    def __init__(self, plugin):
        """
        Constructor of the class, received the "owner"
        plugin as the first argument to be stored for
        latter usage.

        :type plugin: Plugin
        :param plugin: The owner plugin for the test
        object to be created.
        """

        self.plugin = plugin

    def get_bundle(self):
        return ()

    def set_up(self, test_case):
        if hasattr(self.plugin, "system"): test_case.system = self.plugin.system

    def tear_down(self, test_case):
        pass

    def run_all(self, plugin = None, runner = None, verbosity = 1):
        suite = unittest.TestSuite()
        loader = unittest.TestLoader()

        for test_case in self.get_bundle():
            test_case.plugin = plugin
            partial = loader.loadTestsFromTestCase(test_case)
            suite.addTest(partial)

        runner = runner or unittest.TextTestRunner(verbosity = verbosity)
        return runner.run(suite)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from colony.base import exceptions

def verify(condition, message = None, code = None, exception = None, **kwargs):
    if condition: return
    exception = exception or exceptions.AssertionError
    kwargs = dict(kwargs)
    if not message == None: kwargs["message"] = message
    if not code == None: kwargs["code"] = code
    raise exception(**kwargs)

def verify_equal(first, second, message = None, code = None, exception = None, **kwargs):
    message = message or "Expected %s got %s" % (repr(second), repr(first))
    return verify(
        first == second,
        message = message,
        code = code,
        exception = exception,
        **kwargs
    )

def verify_not_equal(first, second, message = None, code = None, exception = None, **kwargs):
    message = message or "Expected %s not equal to %s" % (repr(first), repr(second))
    return verify(
        not first == second,
        message = message,
        code = code,
        exception = exception,
        **kwargs
    )

def verify_type(value, types, null = True, message = None, code = None, exception = None, **kwargs):
    message = message or "Expected %s to have type %s" % (repr(value), repr(types))
    return verify(
        (null and value == None) or isinstance(value, types),
        message = message,
        code = code,
        exception = exception,
        **kwargs
    )

def verify_many(sequence, message = None, code = None, exception = None, **kwargs):
    for condition in sequence:
        verify(
            condition,
            message = message,
            code = code,
            exception = exception,
            **kwargs
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

def version_cmp(version, version_compare):
    """
    Compares the given version string against a compare string
    that may contain wildcard values.

    This way it's possible to compare it against version ranges
    defined in the form of wildcard values.

    :type version: String
    :param version: The base version string to be verified
    by the comparison.
    :type version_compare: String
    :param version_compare: The compare version string to be used
    as reference in the comparison (may contain wildcard values).
    :rtype: bool
    :return: The result of the version string comparison (includes
    wildcard comparison).
    """

    # splits both the (base) version string and the compare string
    # into the three version parts (major, medium and minor) to get
    # them into the normal tuple comparison
    major, medium, minor = version.split(".")
    major_compare, medium_compare, minor_compare = version_compare.split(".")

    # compares all the three parts of the version string so that
    # if there is a mismatch in the version number or in case there
    # is no wildcard present in the compare string the comparison fails
    if not major_compare == "x" and not major == major_compare: return False
    if not medium_compare == "x" and not medium == medium_compare: return False
    if not minor_compare == "x" and not minor == minor_compare: return False

    # returns valid version, all the version tests have passed successfully
    # the version is considered to be valid (equivalent)
    return True

def version_is_concrete(version):
    """
    Verifies if the provided version number is of type concrete
    (all the partial number are specified), or if the version
    string contains wildcard values that match an open range
    of version values.

    :param version: String
    :param version: The string containing the version to be verified
    to be of type concrete.
    :rtype: bool
    :return: The result of the version concrete testing.
    """

    # checks if the wildcard character is present in the version
    # string, if it is the version is considered to be wildcard
    # then returns the negation of that as the is concrete value
    is_wildcard = "x" in version
    return not is_wildcard



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import copy

from colony.base import legacy

class StringBuffer(object):
    """
    The string buffer class, used to provide an
    in-memory file like object for fast access.

    The class is provided as an alternative to the
    pre-defined memory buffers.
    """

    softspace = 0
    """ The soft space value """

    closed = False
    """ The closed value """

    def __init__(self, fast = True, btype = None):
        """
        Constructor of the class.

        :type btype: Type
        :param btype: The default base type to be used for the result
        in case it's not provided a smart operation will try to determine
        the best type of string for the joining.
        :type fast: bool
        :param fast: The fast flag to control the string buffer type, notice
        that using the fast flag has it's toll in terms of features.
        """

        self.string_list = []
        self.current_value = str()
        self.btype = btype
        self.dirty = False
        self.current_position = 0
        self.current_size = 0
        self.fast = fast

        if fast:
            self.write = self._write_fast
        else:
            self.write = self._write_slow

    def read(self, size = None):
        """
        Reads a buffer from the string buffer with the
        given maximum size.

        :type size: int
        :param size: The maximum size of the buffer to be read.
        """

        # regenerates the current value
        self.regenerate()

        if size:
            return_value = self.current_value[self.current_position:self.current_position + size]
            self.seek(size, os.SEEK_CUR)
        else:
            return_value = self.current_value[self.current_position:]
            self.seek(self.current_size, os.SEEK_SET)

        # returns the return value
        return return_value

    def write(self, string_value):
        """
        Writes the given string value to the string buffer.

        The method implementation is a placeholder as runtime
        replacement will be performed at build time.

        :type string_value: String
        :param string_value: The string value to be written.
        """

        pass

    def close(self):
        """
        Closes the string buffer.
        """

        pass

    def flush(self):
        """
        Flushes the string buffer.
        """

        pass

    def reset(self):
        """
        Resets the string buffer.
        """

        self.string_list = []
        self.current_value = str()
        self.dirty = False
        self.current_position = 0
        self.current_size = 0

    def seek(self, offset, whence = os.SEEK_SET):
        """
        Seeks the string buffer to the given offset with the given jump mode,
        defined with the whence.

        :type offset: int
        :param offset: The offset of the jump.
        :type whence: int
        :param whence: The jump mode to be used.
        """

        if whence == os.SEEK_SET:
            self.current_position = offset
        elif whence == os.SEEK_END:
            self.current_position = self.current_size + offset
        elif whence == os.SEEK_CUR:
            if self.current_position + offset < self.current_size:
                self.current_position += offset
            else:
                self.current_position = self.current_size

    def eof(self):
        """
        Returns if the end of file (eof) has been reached.

        :rtype: bool
        :return: If the end of file has been reached.
        """

        return self.current_position == self.current_size

    def next(self):
        """
        Retrieves the next string item from the string buffer.

        :rtype: String
        :return: The next string item from the string buffer.
        """

        return None

    def tell(self):
        """
        Retrieves the current offset position in the string buffer.

        :rtype: int
        :return: The current offset position in the string buffer.
        """

        return self.current_position

    def truncate(self):
        """
        Truncates the string buffer value.
        """

        pass

    def readline(self, size = None):
        """
        Retrieves a line from the string buffer,
        with the given maximum size.

        :type size: int
        :param size: The maximum size of the line to be read.
        :rtype: String
        :return: The read line.
        """

        return None

    def readlines(self, sizehint = None):
        """
        Retrieves a series of lines from the string buffer,
        with the given maximum size.

        :type sizehint: int
        :param sizehint: The maximum size of the lines to be read.
        :rtype: List
        :return: The read lines.
        """

        return None

    def writelines(self, lines):
        """
        Writes the given lines to the string buffer.

        :type lines: List
        :param lines: The lines to be written to the string buffer.
        """

        # iterates over all the lines
        for line in lines:
            # writes the lines to the string buffer
            self.write(line + "\n")

    def isatty(self):
        """
        Returns if the file buffer is of type tty.

        :rtype: bool
        :return: If the file buffer is of type tty.
        """

        return False

    def getvalue(self):
        """
        Retrieves the current string value.
        """

        return self.get_value()

    def get_value(self):
        """
        Retrieves the current string value.
        """

        # regenerates the current value
        self.regenerate()

        # returns the current value
        return self.current_value

    def is_empty(self):
        """
        Returns if the current buffer is empty.

        :rtype: bool
        :return: If the current buffer is empty
        """

        return not self.current_size > 0

    def regenerate(self):
        """
        Regenerates the current value, this is an
        expensive operation and should be performed
        only in extreme situations.
        """

        # in case both the current buffer is not dirty
        # and the fast mode is not enabled the regeneration
        # operation is ignored, control flow returned
        if not self.dirty and not self.fast: return

        # runs the regenerate operation for the current
        # buffer so that the complete set of partial
        # values are joined as a simple buffer value
        self._regenerate()

    def duplicate(self):
        """
        Duplicates the string buffer, returning the
        duplicated value.

        :rtype: StringBuffer
        :return: The duplicated string buffer.
        """

        # creates the new string buffer instance
        duplicated_string_buffer = StringBuffer()

        # sets the duplicated string buffer values
        duplicated_string_buffer.string_list = copy.copy(self.string_list)
        duplicated_string_buffer.current_value = copy.copy(self.current_value)
        duplicated_string_buffer.dirty = self.dirty
        duplicated_string_buffer.current_position = self.current_position
        duplicated_string_buffer.fast = self.fast

        # returns the duplicated string buffer
        return duplicated_string_buffer

    def rollback_last(self, item_count = 1):
        """
        Rollsback the last write.

        :type item_count: int
        :param item_count: The number of items
        to be "rollbacked".
        """

        # iterates over the range of item count
        for _index in range(item_count):
            # pops an item from the string list
            self.string_list.pop()

    def get_last(self, index = -1):
        """
        Retrieves the last write.

        :type index: int
        :param index: The index to retrieve from
        the string list.
        :rtype: String
        :return: The last write.
        """

        # retrieves the absolute index
        absolute_index = abs(index)

        # retrieves the string list length
        string_list_length = len(self.string_list)

        # in case the absolute index "overflows"
        # the string list length
        if absolute_index > string_list_length:
            # return invalid
            return None

        # returns the "last" element
        return self.string_list[index]

    def output_file(self, path):
        """
        Outputs the current file contents to the file
        defined in the target path.

        This method provides a simple way to debug the
        information contained in the buffer.

        :type path: String
        :param path: The path to the file that is going
        to be used to place the buffer contents.
        """

        position = self.tell()
        self.seek(0)
        try:
            contents = self.read()
            file = open(path, "wb")
            try: file.write(contents)
            finally: file.close()
        finally:
            self.seek(position)

    def _write_fast(self, string_value):
        """
        Writes the string value in fast mode.

        :type string_value: String
        :param string_value: The string value to be written.
        """

        self.string_list.append(string_value)

    def _write_slow(self, string_value):
        """
        Writes the string value in slow mode.

        :type string_value: String
        :param string_value: The string value to be written.
        """

        self.string_list.append(string_value)
        self.dirty = True
        self.current_size += len(string_value)
        self.current_position = self.current_size

    def _regenerate(self):
        """
        Regenerates the current value (auxiliary method).

        This may be a very expensive method as it re-joins
        the complete set of parts of the current buffer.
        """

        # retrieves the base string or bytes value and uses
        # it to join the complete set of string values that
        # are part of the current buffer
        base = self._base_type()
        self.current_value = base.join(self.string_list)

        # recreates the string list with the current value
        self.string_list = [self.current_value]

        # unsets the dirty flag, so that the re-generate
        # operation is marked as not required until new
        # data is added to the current string buffer
        self.dirty = False

    def _base_type(self):
        """
        Determines the base string type value that may be used
        to join the various string value parts of the buffer.

        This is required so that no invalid types are used causing
        an exception to be raised in the joining.

        In case the current python interpreter does not allow
        byte buffers or in case the base type is fixed in the
        constructor, this method returns immediately.

        :rtype: String
        :return: The string value that may be used as the basis for
        the joining of the various components of the string list/buffer.
        """

        if not legacy.PYTHON_3: return ""
        if self.btype == str: return ""
        if self.btype == legacy.BYTES: return b""
        if self.btype == legacy.UNICODE: return legacy.u("")
        is_bytes = True
        for value in self.string_list:
            if not type(value) == str: continue
            is_bytes = False
            break
        return b"" if is_bytes else ""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import time

from colony.base import legacy

DEFAULT_NUMBER_RETRIES = 3
""" The default number of retries to register or unregister """

DEFAULT_RETRY_SLEEP = 1
""" The default sleep time between retries """

def execute_retries(
    callable,
    number_retries = DEFAULT_NUMBER_RETRIES,
    retry_sleep = DEFAULT_RETRY_SLEEP
):
    """
    Executes the given callable retrying the call in case an exception occurs.
    The number of retries and the time between retries is configurable.
    The method returns the return value from the call or raises the last
    known exception.

    :type callable: Callable
    :param callable: The callable to be called using retries.
    :type number_retries: int
    :param number_retries: The number of retries to be used.
    :type retry_sleep: int´
    :param retry_sleep: The sleep time between retries.
    :rtype: Object
    :return: The return value from the callable.
    """

    # iterates over the range of the number retries (plus one)
    # the last iteration is used for exception re-raising
    for index in range(number_retries + 1):
        try:
            # calls the callable object, retrieving
            # and saving the return value
            return_value = callable()

            # breaks the loop, because there is
            # no exception raised (successful call)
            break
        except Exception:
            # in case it's the last index position
            # the exception should be re-raised
            if index == number_retries: raise

            # sleeps a while to avoid problems
            time.sleep(retry_sleep)

    # returns the callable return value
    return return_value

def call_safe(callable, *args, **kwargs):
    """
    Method used to call a callable object using a "safe" approach,
    meaning that each of its keyword arguments will be validated
    for existence in the target callable definition.

    In case the validation of the keyword argument fails the same
    argument is removed from the map of keyword arguments.

    Note that in case the wildcard based kwargs value exists in
    the callable definition the callable is immediately considered
    to be valid and the call is ran.

    :type callable: Callable
    :param callable: The callable that is going to have the keyword
    based arguments validated and the get called.
    :rtype: object
    :return: The resulting value from the safe call of the provided
    callable, this may have any data type.
    """

    # retrieves the arguments specification to the provided callable
    # and retrieves the various argument names and the existence or
    # not of the wildcard kwargs value in the callable and in case it
    # exists runs the callable call immediately
    argspec = legacy.getargspec(callable)
    method_args = argspec[0]
    method_kwargs = argspec[2]
    if method_kwargs: return callable(*args, **kwargs)

    # iterates over the complete set of keyword based arguments to be
    # used in the call and validates them against the method specification
    # in case they do not exist in the specification deletes them from
    # the map of keyword based arguments (not going to be sent)
    for name in legacy.keys(kwargs):
        if name in method_args: continue
        del kwargs[name]

    # runs the callable with the "remaining" arguments and keyword arguments
    # returning the value to the caller method
    return callable(*args, **kwargs)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import sys

VERSION = "%(version)s"
""" The version value, that identifies the current
running version, this value should have three components """

RELEASE = "%(release)s"
""" The release value identifying the continuous integration
based value, this value should be considered internal """

BUILD = "%(build)s"
""" The build value, considered to be internal """

RELEASE_DATE = "%(date)s"
""" The release date value conformant with the typical
data and string values """

RELEASE_DATE_TIME = "%(date_time)s"
""" The release date time value that should timestamp the
current system """

ENVIRONMENT_VERSION = str(sys.version_info[0]) + "." + str(sys.version_info[1]) + "." + str(sys.version_info[2]) + "-" + str(sys.version_info[3])
""" The environment version, constructed using information
from the currently running python interpreter """

ENVIRONMENT = "python-" + sys.platform + " " + ENVIRONMENT_VERSION
""" The environment string that should identify in a minimal
way the currently running system/environment """

__GENERATED__ = "%(generated)s"
""" Flag based global object that controls if the
values in the current module have been already
generated by a previous call or tool """

# % END_INFORMATION % #

DEFAULT_ENCODING = "utf-8"
""" The default encoding value to be used to convert
the information values into string, this value should
be set as compatible with the information file encoding
so that no encoding related exceptions are raised """

DATE_FORMAT = "%b %d %Y"
""" The format used to convert dates to strings """

DATE_TIME_FORMAT = "%b %d %Y %H:%M:%S"
""" The format used to convert date times to strings """

INFORMATION_PATH = "../res/colony.json"
""" The relative path to the file containing the
complete "release" information about the current
colony implementation running """

def generate(information = None, encoding = DEFAULT_ENCODING):
    # sets the various environment variables as the
    # global reference to their values to avoid
    # confusing local allocation
    global VERSION
    global RELEASE
    global BUILD
    global RELEASE_DATE
    global RELEASE_DATE_TIME
    global __GENERATED__

    # checks if the values in the current module have been
    # already generated in such case there's no need to generate
    # them again (expensive operation)
    if __GENERATED__ == "1": return

    # sets the generated flag to indicate that the values
    # in the module have been already processed
    __GENERATED__ = "1"

    # in case no information is provided the configuration
    # file must be uses to retrieve the proper information
    # for the global values, this approach has impact in
    # the performance of the generation and also requires
    # the usage of the JSON module (may not exist)
    if information == None:
        # tries to import the JSON module in case of failure
        # a silent return must occur to avoid problems
        try: import os, json, datetime
        except ImportError: return

        # retrieves the current directory path and uses it to
        # infer the location of the information file to the
        # current colony instance, then uses that path to load
        # the associated JSON information
        directory = os.path.dirname(__file__)
        file_path = os.path.join(directory, INFORMATION_PATH)
        file_path = os.path.normpath(file_path)
        file = open(file_path, "rb")
        try: data = file.read()
        finally: file.close()
        data = data.decode(encoding)
        information = json.loads(data)

        # retrieves the current datetime and formats it
        # according to the two pre-defined formats
        current_datetime = datetime.datetime.utcnow()
        current_date_string = current_datetime.strftime(DATE_FORMAT)
        current_date_time_string = current_datetime.strftime(DATE_TIME_FORMAT)

        # sets both the date string and the date time string
        # in the symbols map for template reference
        information["date"] = current_date_string
        information["date_time"] = current_date_time_string

    # processes the various items of the information
    # module using the provided information map, this
    # map must contain all the information required
    # for the generation of these string, otherwise
    # this function will fails in exception
    VERSION = VERSION % information
    RELEASE = RELEASE % information
    BUILD = BUILD % information
    RELEASE_DATE = RELEASE_DATE % information
    RELEASE_DATE_TIME = RELEASE_DATE_TIME % information

# runs the generate task at import time to provide
# the most update values to the importer agent
generate()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

from . import number_util

COEFFICIENT_VALUES = (
    2, 3, 4, 5, 6, 7, 8, 9
)
""" The tuple containing the coefficient values """

def calculate_tax_number_control_value(tax_number):
    """
    Calculates the control value for the tax number.

    :type tax_number: int
    :param tax_number: The tax number to calculate the
    control value.
    :rtype: int
    :return: The calculated control value.
    """

    # calculates the number of digits in the number
    number_digits = number_util.get_number_length(tax_number)

    # in case the number of digits is invalid  must raise an
    # exception indicating the error, note that this requirement
    # in terms of size ensures that no tax number may start with
    # a number like zero (reducing the length of the valid digits)
    if not number_digits == 8:
        raise RuntimeError("Invalid tax number length '%d'" % number_digits)

    # calculates the control value for the number and then
    # returns it to the caller method
    control_value = _calculate_control_value(tax_number)
    return control_value

def calculate_id_number_control_value(id_number):
    """
    Calculates the control value for the id number.

    :type id_number: int
    :param id_number: The id number to calculate the
    control value.
    :rtype: int
    :return: The calculated control value.
    """

    # calculates the number of digits in the number
    number_digits = number_util.get_number_length(id_number)

    # in case the number of digits is invalid
    # must raise an exception indicating the error
    if number_digits < 6 or number_digits > 8:
        raise RuntimeError("Invalid id number length '%d'" % number_digits)

    # adds zeros to the right of the number (if necessary)
    id_number *= 10 ** (8 - number_digits)

    # calculates the control value for the number and then
    # returns it to the caller method
    control_value = _calculate_control_value(id_number)
    return control_value

def _calculate_control_value(number):
    """
    Calculates the control value for the given number, this
    control value is related with the calculus of a tax
    number number or id number.

    :type number: int
    :param number: The number to calculate the control value,
    the length of this number is not verifies and may assume
    any size, as that's not part of the control value calculus.
    :rtype: int
    :return: The control value for the given number.
    """

    # calculates the number of digits in the number
    number_digits = number_util.get_number_length(number)

    # starts the accumulator
    accumulator = 0

    # iterates over the range of the number of digits
    for index in range(number_digits):
        # retrieves the current digit and then the associated
        # coefficient value, incrementing then the accumulator
        # with the current digit multiplied with the current
        # coefficient (as expected by algorithm)
        current_digit = number_util.get_digit(number, index)
        current_coefficient = COEFFICIENT_VALUES[index]
        accumulator += current_digit * current_coefficient

    # calculates the partial value from
    # the accumulator value
    partial_value = accumulator % 11

    # in case the partial value is zero
    # or one sets the control value as zero
    if partial_value in (0, 1): control_value = 0

    # otherwise calculates the control value based
    # on the partial value
    else: control_value = 11 - partial_value

    # returns the control value
    return control_value



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt> & Tiago Silva <tsilva@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import re

UNDERSCORE_FIRST_VALUE = "(.)([A-Z][a-z]+)"
""" The underscore first value """

UNDERSCORE_FIRST_REGEX = re.compile(UNDERSCORE_FIRST_VALUE)
""" The underscore first regex """

UNDERSCORE_SECOND_VALUE = "([a-z0-9])([A-Z])"
""" The underscore second value """

UNDERSCORE_SECOND_REGEX = re.compile(UNDERSCORE_SECOND_VALUE)
""" The underscore second regex """

def xor_string_value(first_string, second_string):
    """
    Runs the xor bitwise operation over all the items
    of both strings, retrieving the result.

    :type first_string: String
    :param first_string: The first string to the xor operation.
    :type second_string: String
    :param second_string: The second string to the xor operation.
    :rtype: String
    :return: The "string" result of the xor operation.
    """

    # retrieves the length of the first string
    first_string_length = len(first_string)

    # retrieves the length of the second string
    second_string_length = len(second_string)

    # in case the string have different length
    if not first_string_length == second_string_length:
        # raises the value erorr exception
        raise ValueError("Arguments to xor string must have the same length")

    # creates the list to hold the resulting characters
    result_character_list = []

    # iterates over all the characters in both strings
    for first_character, second_character in zip(first_string, second_string):
        # calculates the result character from the first and second characters
        result_character = chr(ord(first_character) ^ ord(second_character))

        # adds the result character to the result character list
        result_character_list.append(result_character)

    # joins the result character list to retrieve
    # the final xor result
    xor_result = "".join(result_character_list)

    # returns the xor result
    return xor_result

def to_underscore(string_value):
    """
    Converts the given camel cased string value into
    the underscore notation.
    This method is useful to treat class string values in python.

    :type string_value: String
    :param string_value: The camel cased string value to be converted
    into underscore notation.
    :rtype: String
    :return: The converted underscore notation string value.
    """

    # converts the string value into the initial underscore notation
    string_value_underscore = UNDERSCORE_FIRST_REGEX.sub(r"\1_\2", string_value)

    # converts the string value underscore into the next underscore notation
    string_value_underscore = UNDERSCORE_SECOND_REGEX.sub(r"\1_\2", string_value_underscore)

    # converts the string value in underscore notation to lowercase
    string_value_underscore = string_value_underscore.lower()

    # returns the string value in underscore notation
    return string_value_underscore

def to_camelcase(string_value):
    """
    Converts the given underscore notation string value into
    the camel case notation.
    This method is useful to treat class string values in python.

    :type string_value: String
    :param string_value: The underscore notation string
    value to be converted into camel case notation.
    :rtype: String
    :return: The converted camel case notation string value.
    """

    # splits the string value retrieving
    # the string value tokens
    string_value_tokens = string_value.split("_")

    # creates a list with all the string value tokens capitalized
    string_value_tokens_capitalized = [value.capitalize() for value in string_value_tokens]

    # joins the various string value tokens
    # creating the string value camel case
    string_value_camelcase = "".join(string_value_tokens_capitalized)

    # returns the string value camel case
    return string_value_camelcase

def pluralize(string_value):
    """
    Pluralizes the given string value, using a
    typical heuristic for it.

    :type string_value: String
    :param string_value: The string to be pluralized.
    :rtype: String
    :return: The pluralized string.
    """

    return "%ss" % string_value

def capitalize_all(string_value):
    """
    Takes a space separated string value
    and capitalizes all of its words.

    :type string_value: String
    :param string_value: The space separated string
    to be capitalized.
    :rtype: String
    :return: The capitalized string.
    """

    # splits the string value retrieving
    # the string value tokens
    string_value_tokens = string_value.split()

    # creates a list with all the string value tokens capitalized
    string_value_tokens_capitalized = [value.capitalize() for value in string_value_tokens]

    # joins the various string value tokens
    # creating the string value capitalized
    string_value_capitalized = " ".join(string_value_tokens_capitalized)

    # returns the string value capitalized
    return string_value_capitalized

def join(first_value, second_value, join_value = "/"):
    """
    Joins two string making sure that the join value is
    not repeated in the "join point" (no double join).
    The "join point" is guaranteed to be unique.

    :type first_value: String
    :param first_value: The first string to be joined.
    :type second_value: String
    :param second_value: The second string to be joined.
    :type join_value: String
    :param join_value: The join value to be verifies and
    used in the "join point".
    :rtype: String
    :return: The joined string ensured to be with a single
    join point (not repeated).
    """

    # removes the join value from both the first and second value
    # to avoid duplicate values in the joining
    first_value = not first_value.endswith(join_value) and first_value or first_value[:-1]
    second_value = not second_value.startswith(join_value) and second_value or second_value[1:]

    # joins the first and second values using the join
    # value from the join point
    string_value = join_value.join((first_value, second_value))

    # returns the "resulting" stirng value
    return string_value



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

COUNTRIES = (
    "Afghanistan",
    "Albania",
    "Algeria",
    "Andorra",
    "Angola",
    "Antigua & Deps",
    "Argentina",
    "Armenia",
    "Australia",
    "Austria",
    "Azerbaijan",
    "Bahamas",
    "Bahrain",
    "Bangladesh",
    "Barbados",
    "Belarus",
    "Belgium",
    "Belize",
    "Benin",
    "Bhutan",
    "Bolivia",
    "Bosnia Herzegovina",
    "Botswana",
    "Brazil",
    "Brunei",
    "Bulgaria",
    "Burkina",
    "Burundi",
    "Cambodia",
    "Cameroon",
    "Canada",
    "Cape Verde",
    "Central African Rep",
    "Chad",
    "Chile",
    "China",
    "Colombia",
    "Comoros",
    "Congo",
    "Congo Democratic Republic",
    "Costa Rica",
    "Croatia",
    "Cuba",
    "Cyprus",
    "Czech Republic",
    "Denmark",
    "Djibouti",
    "Dominica",
    "Dominican Republic",
    "East Timor",
    "Ecuador",
    "Egypt",
    "El Salvador",
    "Equatorial Guinea",
    "Eritrea",
    "Estonia",
    "Ethiopia",
    "Fiji",
    "Finland",
    "France",
    "Gabon",
    "Gambia",
    "Georgia",
    "Germany",
    "Ghana",
    "Greece",
    "Grenada",
    "Guatemala",
    "Guinea",
    "Guinea-Bissau",
    "Guyana",
    "Haiti",
    "Honduras",
    "Hong Kong",
    "Hungary",
    "Iceland",
    "India",
    "Indonesia",
    "Iran",
    "Iraq",
    "Ireland",
    "Israel",
    "Italy",
    "Ivory Coast",
    "Jamaica",
    "Japan",
    "Jordan",
    "Kazakhstan",
    "Kenya",
    "Kiribati",
    "Korea North",
    "Korea South",
    "Kosovo",
    "Kuwait",
    "Kyrgyzstan",
    "Laos",
    "Latvia",
    "Lebanon",
    "Lesotho",
    "Liberia",
    "Libya",
    "Liechtenstein",
    "Lithuania",
    "Luxembourg",
    "Macao",
    "Macedonia",
    "Madagascar",
    "Malawi",
    "Malaysia",
    "Maldives",
    "Mali",
    "Malta",
    "Marshall Islands",
    "Mauritania",
    "Mauritius",
    "Mexico",
    "Micronesia",
    "Moldova",
    "Monaco",
    "Mongolia",
    "Montenegro",
    "Morocco",
    "Mozambique",
    "Myanmar",
    "Namibia",
    "Nauru",
    "Nepal",
    "Netherlands",
    "New Zealand",
    "Nicaragua",
    "Niger",
    "Nigeria",
    "Norway",
    "Oman",
    "Pakistan",
    "Palau",
    "Panama",
    "Papua New Guinea",
    "Paraguay",
    "Peru",
    "Philippines",
    "Poland",
    "Portugal",
    "Qatar",
    "Romania",
    "Russia",
    "Rwanda",
    "St Kitts & Nevis",
    "St Lucia",
    "Saint Vincent & the Grenadines",
    "Samoa",
    "San Marino",
    "Sao Tome & Principe",
    "Saudi Arabia",
    "Senegal",
    "Serbia",
    "Seychelles",
    "Sierra Leone",
    "Singapore",
    "Slovakia",
    "Slovenia",
    "Solomon Islands",
    "Somalia",
    "South Africa",
    "Spain",
    "Sri Lanka",
    "Sudan",
    "Suriname",
    "Swaziland",
    "Sweden",
    "Switzerland",
    "Syria",
    "Taiwan",
    "Tajikistan",
    "Tanzania",
    "Thailand",
    "Togo",
    "Tonga",
    "Trinidad & Tobago",
    "Tunisia",
    "Turkey",
    "Turkmenistan",
    "Tuvalu",
    "Uganda",
    "Ukraine",
    "United Arab Emirates",
    "United Kingdom",
    "United States",
    "Uruguay",
    "Uzbekistan",
    "Vanuatu",
    "Vatican City",
    "Venezuela",
    "Vietnam",
    "Yemen",
    "Zambia",
    "Zimbabwe"
)
""" The complete set of countries using the colony's default
descriptive name so that it may be used in locale based environment
note that these names do not conform with any standard """

COUNTRIES_ISO = {
    "afghanistan" : ("AF", "AFG", "004"),
    "albania" : ("AL", "ALB", "008"),
    "algeria" : ("DZ", "DZA", "012"),
    "andorra" : ("AD", "AND", "020"),
    "angola" : ("AO", "AGO", "024"),
    "antigua & deps" : ("AG", "ATG", "028"),
    "argentina" : ("AR", "ARG", "032"),
    "armenia" : ("AM", "ARM", "051"),
    "australia" : ("AU", "AUS", "036"),
    "austria" : ("AT", "AUT", "040"),
    "azerbaijan" : ("AZ", "AZE", "031"),
    "bahamas" : ("BS", "BHS", "044"),
    "bahrain" : ("BH", "BHR", "048"),
    "bangladesh" : ("BD", "BGD", "050"),
    "barbados" : ("BB", "BRB", "052"),
    "belarus" : ("BY", "BLR", "112"),
    "belgium" : ("BE", "BEL", "056"),
    "belize" : ("BZ", "BLZ", "084"),
    "benin" : ("BJ", "BEN", "204"),
    "bhutan" : ("BT", "BTN", "064"),
    "bolivia" : ("BO", "BOL", "068"),
    "bosnia herzegovina" : ("BA", "BIH", "070"),
    "botswana" : ("BW", "BWA", "072"),
    "brazil" : ("BR", "BRA", "076"),
    "brunei" : ("BN", "BRN", "096"),
    "bulgaria" : ("BG", "BGR", "100"),
    "burkina" : ("BF", "BFA", "854"),
    "burundi" : ("BI", "BDI", "108"),
    "cambodia" : ("KH", "KHM", "116"),
    "cameroon" : ("CM", "CMR", "120"),
    "canada" : ("CA", "CAN", "124"),
    "cape verde" : ("CV", "CPV", "132"),
    "central african rep" : ("CF", "CAF", "140"),
    "chad" : ("TD", "TCD", "148"),
    "chile" : ("CL", "CHL", "152"),
    "china" : ("CN", "CHN", "156"),
    "colombia" : ("CO", "COL", "170"),
    "comoros" : ("KM", "COM", "174"),
    "congo" : ("CG", "COG", "178"),
    "congo democratic republic" : ("CD", "COD", "180"),
    "costa rica" : ("CR", "CRI", "188"),
    "croatia" : ("HR", "HRV", "191"),
    "cuba" : ("CU", "CUB", "192"),
    "cyprus" : ("CY", "CYP", "196"),
    "czech republic" : ("CZ", "CZE", "203"),
    "denmark" : ("DK", "DNK", "208"),
    "djibouti" : ("DJ", "DJI", "262"),
    "dominica" : ("DM", "DMA", "212"),
    "dominican republic" : ("DO", "DOM", "214"),
    "east timor" : ("TL", "TLS", "626"),
    "ecuador" : ("EC", "ECU", "218"),
    "egypt" : ("EG", "EGY", "818"),
    "el salvador" : ("SV", "SLV", "222"),
    "equatorial guinea" : ("GQ", "GNQ", "226"),
    "eritrea" : ("ER", "ERI", "232"),
    "estonia" : ("EE", "EST", "233"),
    "ethiopia" : ("ET", "ETH", "231"),
    "fiji" : ("FJ", "FJI", "242"),
    "finland" : ("FI", "FIN", "246"),
    "france" : ("FR", "FRA", "250"),
    "gabon" : ("GA", "GAB", "266"),
    "gambia" : ("GM", "GMB", "270"),
    "georgia" : ("GE", "GEO", "268"),
    "germany" : ("DE", "DEU", "276"),
    "ghana" : ("GH", "GHA", "288"),
    "greece" : ("GR", "GRC", "300"),
    "grenada" : ("GD", "GRD", "308"),
    "guatemala" : ("GT", "GTM", "320"),
    "guinea" : ("GN", "GIN", "324"),
    "guinea-bissau" : ("GW", "GNB", "624"),
    "guyana" : ("GY", "GUY", "328"),
    "haiti" : ("HT", "HTI", "332"),
    "honduras" : ("HN", "HND", "340"),
    "hong kong" : ("HK", "HKG", "344"),
    "hungary" : ("HU", "HUN", "348"),
    "iceland" : ("IS", "ISL", "352"),
    "india" : ("IN", "IND", "356"),
    "indonesia" : ("ID", "IDN", "360"),
    "iran" : ("IR", "IRN", "364"),
    "iraq" : ("IQ", "IRQ", "368"),
    "ireland" : ("IE", "IRL", "372"),
    "israel" : ("IL", "ISR", "376"),
    "italy" : ("IT", "ITA", "380"),
    "ivory coast" : ("CI", "CIV", "384"),
    "jamaica" : ("JM", "JAM", "388"),
    "japan" : ("JP", "JPN", "392"),
    "jordan" : ("JO", "JOR", "400"),
    "kazakhstan" : ("KZ", "KAZ", "398"),
    "kenya" : ("KE", "KEN", "404"),
    "kiribati" : ("KI", "KIR", "296"),
    "korea north" : ("KP", "PRK", "408"),
    "korea south" : ("KR", "KOR", "410"),
    "kosovo" : ("KO", "KOS", "780"),
    "kuwait" : ("KW", "KWT", "414"),
    "kyrgyzstan" : ("KG", "KGZ", "417"),
    "laos" : ("LA", "LAO", "418"),
    "latvia" : ("LV", "LVA", "428"),
    "lebanon" : ("LB", "LBN", "422"),
    "lesotho" : ("LS", "LSO", "426"),
    "liberia" : ("LR", "LBR", "430"),
    "libya" : ("LY", "LBY", "434"),
    "liechtenstein" : ("LI", "LIE", "438"),
    "lithuania" : ("LT", "LTU", "440"),
    "luxembourg" : ("LU", "LUX", "442"),
    "macao" : ("MO", "MAC", "446"),
    "macedonia" : ("MK", "MKD", "807"),
    "madagascar" : ("MG", "MDG", "450"),
    "malawi" : ("MW", "MWI", "454"),
    "malaysia" : ("MY", "MYS", "458"),
    "maldives" : ("MV", "MDV", "462"),
    "mali" : ("ML", "MLI", "466"),
    "malta" : ("MT", "MLT", "470"),
    "marshall islands" : ("MH", "MHL", "584"),
    "mauritania" : ("MR", "MRT", "478"),
    "mauritius" : ("MU", "MUS", "480"),
    "mexico" : ("MX", "MEX", "484"),
    "micronesia" : ("FM", "FSM", "583"),
    "moldova" : ("MD", "MDA", "498"),
    "monaco" : ("MC", "MCO", "492"),
    "mongolia" : ("MN", "MNG", "496"),
    "montenegro" : ("ME", "MNE", "499"),
    "morocco" : ("MA", "MAR", "504"),
    "mozambique" : ("MZ", "MOZ", "508"),
    "myanmar" : ("MM", "MMR", "104"),
    "namibia" : ("NA", "NAM", "516"),
    "nauru" : ("NR", "NRU", "520"),
    "nepal" : ("NP", "NPL", "524"),
    "netherlands" : ("NL", "NLD", "528"),
    "new zealand" : ("NZ", "NZL", "554"),
    "nicaragua" : ("NI", "NIC", "558"),
    "niger" : ("NE", "NER", "562"),
    "nigeria" : ("NG", "NGA", "566"),
    "norway" : ("NO", "NOR", "578"),
    "oman" : ("OM", "OMN", "512"),
    "pakistan" : ("PK", "PAK", "586"),
    "palau" : ("PW", "PLW", "585"),
    "panama" : ("PA", "PAN", "591"),
    "papua new guinea" : ("PG", "PNG", "598"),
    "paraguay" : ("PY", "PRY", "600"),
    "peru" : ("PE", "PER", "604"),
    "philippines" : ("PH", "PHL", "608"),
    "poland" : ("PL", "POL", "616"),
    "portugal" : ("PT", "PRT", "620"),
    "qatar" : ("QA", "QAT", "634"),
    "romania" : ("RO", "ROU", "642"),
    "russia" : ("RU", "RUS", "643"),
    "rwanda" : ("RW", "RWA", "646"),
    "st kitts & nevis" : ("KN", "KNA", "659"),
    "st lucia" : ("LC", "LCA", "662"),
    "saint vincent & the grenadines" : ("VC", "VCT", "670"),
    "samoa" : ("WS", "WSM", "882"),
    "san marino" : ("SM", "SMR", "674"),
    "sao tome & principe" : ("ST", "STP", "678"),
    "saudi arabia" : ("SA", "SAU", "682"),
    "senegal" : ("SN", "SEN", "686"),
    "serbia" : ("RS", "SRB", "688"),
    "seychelles" : ("SC", "SYC", "690"),
    "sierra leone" : ("SL", "SLE", "694"),
    "singapore" : ("SG", "SGP", "702"),
    "slovakia" : ("SK", "SVK", "703"),
    "slovenia" : ("SI", "SVN", "705"),
    "solomon islands" : ("SB", "SLB", "090"),
    "somalia" : ("SO", "SOM", "706"),
    "south africa" : ("ZA", "ZAF", "710"),
    "spain" : ("ES", "ESP", "724"),
    "sri lanka" : ("LK", "LKA", "144"),
    "sudan" : ("SD", "SDN", "736"),
    "suriname" : ("SR", "SUR", "740"),
    "swaziland" : ("SZ", "SWZ", "748"),
    "sweden" : ("SE", "SWE", "752"),
    "switzerland" : ("CH", "CHE", "756"),
    "syria" : ("SY", "SYR", "760"),
    "taiwan" : ("TW", "TWN", "158"),
    "tajikistan" : ("TJ", "TJK", "762"),
    "tanzania" : ("TZ", "TZA", "834"),
    "thailand" : ("TH", "THA", "764"),
    "togo" : ("TG", "TGO", "768"),
    "tonga" : ("TO", "TON", "776"),
    "trinidad & tobago" : ("TT", "TTO", "780"),
    "tunisia" : ("TN", "TUN", "788"),
    "turkey" : ("TR", "TUR", "792"),
    "turkmenistan" : ("TM", "TKM", "795"),
    "tuvalu" : ("TV", "TUV", "798"),
    "uganda" : ("UG", "UGA", "800"),
    "ukraine" : ("UA", "UKR", "804"),
    "united arab emirates" : ("AE", "ARE", "784"),
    "united kingdom" : ("GB", "GBR", "826"),
    "united states" : ("US", "USA", "840"),
    "uruguay" : ("UY", "URY", "858"),
    "uzbekistan" : ("UZ", "UZB", "860"),
    "vanuatu" : ("VU", "VUT", "548"),
    "vatican city" : ("VA", "VAT", "336"),
    "venezuela" : ("VE", "VEN", "862"),
    "vietnam" : ("VN", "VNM", "704"),
    "yemen" : ("YE", "YEM", "887"),
    "zambia" : ("ZM", "ZMB", "894"),
    "zimbabwe" : ("ZW", "ZWE", "716")
}
""" The map associating the various countries with the
tuple containing the iso 3166 information (two character
code, three character code and the number code) """

def country_get(name, relaxed = True):
    """
    Retrieves a tuple containing the complete set of ISO 3166
    information for the country with the provided name.

    The returned tuple contains a two character code a three
    character code and the number code.

    An optional relaxed flag allows an automatic conversion of
    the country name to a lower cased version for easy retrieval.

    :type name: String
    :param name: The name of the country for which the information
    is meant to be retrieved.
    :type relaxed: bool
    :param relaxed: If the (country) name retrieval should take a
    relaxed approach meaning that it will be first converted into
    a lower cased version to avoid collisions.
    :rtype: Tuple
    :return: A tuple containing the iso 3166 information for the
    request country.
    :see: http://en.wikipedia.org/wiki/ISO_3166
    """

    if relaxed: name = name.lower()
    return COUNTRIES_ISO.get(name, (None, None, None))



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import os
import tempfile
import threading

from colony.base import legacy

from . import list_util
from . import path_util

PATH_TUPLE_PROCESS_METHOD_PREFIX = "_process_path_tuple_"
""" The prefix to the path tuple process method """

ADD_OPERATION = "add"
""" The add operation """

ADD_NO_REPLACE_OPERATION = "add_no_replace"
""" The add no replace operation """

REMOVE_OPERATION = "remove"
""" The remove operation """

class FileRotator(object):
    """
    Class for handling of writing in files
    with a rotating capability.
    """

    base_file_path = None
    """ The base file path """

    maximum_file_size = None
    """ The maximum file size """

    file_count = None
    """ The number of file to maintained """

    current_file = None
    """ The current file in use """

    current_file_size = None
    """ The size of the current file in use """

    closed = False
    """ The flag that controls the current status of the file rotator """

    def __init__(self, base_file_path, maximum_file_size = 1048576, file_count = 5):
        """
        Constructor of the class.

        :type base_file_path: String
        :param base_file_path: The base file path to be used.
        :type maximum_file_size: int
        :param maximum_file_size: The maximum file size.
        :type file_count: int
        :param file_count: The number of files to be used.
        """

        self.base_file_path = base_file_path
        self.maximum_file_size = maximum_file_size
        self.file_count = file_count

    def open(self):
        """
        Opens the file rotator.
        """

        # starts the rotator
        self._sart_rotator()

        # updates the closed status
        self.closed = False

    def close(self):
        """
        Closes the file rotator.
        """

        # stops the rotator
        self._stop_rotator()

        # updates the closed status
        self.closed = True

    def write(self, string_value, flush = True, encoding = "utf-8"):
        """
        Writes the given string value using
        the current file rotator.

        :type string_value: String
        :param string_value: The string value to be
        written.
        :type flush: bool
        :param flush: If the current file rotator should
        be flushed.
        """

        # retrieves the string value length
        string_value_length = len(string_value)

        # in case the string value overflow the current maximum file size
        if self.current_file_size + string_value_length > self.maximum_file_size:
            # updates the rotator
            self._update_rotator()

        # verifies if the provided string value is "decoded" as unicode
        # if that's the case runs the encoding process using the value
        # of the encoding attribute as the reference for the encoding
        is_unicode = type(string_value) == legacy.UNICODE
        if is_unicode: string_value = string_value.encode(encoding)

        # writes the string value to the current file, note that the
        # value that is being written is already encoded as bytes
        self.current_file.write(string_value)

        # flushes the data in the current file
        flush and self.current_file.flush()

        # increments the current file size with
        # the string value length
        self.current_file_size += string_value_length

    def is_closed(self):
        """
        Returns the current closed status
        for the internal file reference.

        :rtype: bool
        :return: The current closed status
        for the internal file reference.
        """

        return self.closed

    def _sart_rotator(self):
        """
        Starts the file rotator.
        """

        # opens the current file
        self._open_current_file()

    def _stop_rotator(self):
        """
        Stops the file rotator.
        """

        # closes the current file (in case it's open)
        self.current_file and self._close_current_file()

    def _open_current_file(self):
        """
        Opens the current file to be used.
        """

        # opens the current file
        self.current_file = open(self.base_file_path, "ab")

        # seeks to the end of the current file
        self.current_file.seek(0, os.SEEK_END)

        # sets the initial current file size
        self.current_file_size = self.current_file.tell()

    def _close_current_file(self, rename = False):
        """
        Closes the current file being used.
        In case the renaming flag is set the file is
        renamed to the first file in the rotator.

        :type rename: bool
        :param rename: If the renaming on the
        file should be done.
        """

        # flushes the current file
        self.current_file.flush()

        # closes the current file
        self.current_file.close()

        # in case the rename flag is not set
        if not rename:
            # returns immediately
            return

        # creates the target file path for the
        # base file path
        target_file_path = self.base_file_path + ".1"

        # renames the file with to the first incremental index
        os.rename(self.base_file_path, target_file_path)

    def _update_rotator(self):
        """
        Updates the current file rotator state.
        This procedure involves the renaming the current
        set of files in the rotator.
        It also removes the extra file that overflow the
        current file count.
        """

        # retrieves the file count range and runs the
        # reverse operation in it to obtains the correct
        # range value to be used in the rotator operation
        file_count_range = legacy.eager(range(1, self.file_count + 1))
        file_count_range.reverse()

        # iterates over the range of the
        # file count
        for index in file_count_range:
            # creates the target file path from the base
            # file path and the index string
            target_file_path = self.base_file_path + "." + str(index)

            # in case the target file path does not exist
            # there's nothing to be done, skips the loop
            if not os.path.exists(target_file_path): continue

            # in case the index is small than
            # the file count
            if index < self.file_count:
                # renames the file with an incremented index
                os.rename(target_file_path, self.base_file_path + "." + str(index + 1))
            # otherwise (overflow occurred)
            else:
                # removes the target file
                os.remove(target_file_path)

        # closes the current file
        self.current_file and self._close_current_file(True)

        # opens the current file
        self._open_current_file()

class FileContext(object):
    """
    The file context class used to read and write
    contents from files.
    """

    def __init__(self):
        """
        Constructor of the class.
        """

        pass

    def resolve_file_path(self, file_path):
        """
        Resolves the given file path, for the current
        environment it is the same value.

        :type file_path: String
        :param file_path: The file path to be resolved.
        :rtype: String
        :return: The resolved (real) file path.
        """

        # returns the file path (same value)
        return file_path

    def exists_file_path(self, file_path):
        """
        Tests if the given file path exists in
        the current environment.

        :type file_path: String
        :param file_path: The file path to be tested
        for existence.
        :rtype: bool
        :return: The result of the test for
        file existence.
        """

        # checks if the file path exists
        exists_file_path = os.path.exists(file_path)

        # returns the exists file path result
        return exists_file_path

    def is_directory_path(self, file_path):
        """
        Tests if the given file path refers a directory path in
        the current environment.

        :type file_path: String
        :param file_path: The file path to be tested
        for directory referral.
        :rtype: bool
        :return: The result of the test for
        directory referral.
        """

        # tests if the file path refers a directory path
        is_directory_path = os.path.isdir(file_path)

        # returns the is directory path result
        return is_directory_path

    def read_file(self, file_path):
        """
        Reads the contents from the given
        file path.

        :type file_path: String
        :param file_path: The path to the file
        to read from.
        :rtype: String
        :return: The (read) contents from the file.
        """

        # open the file
        file = open(file_path, "rb")

        try:
            # reads the file contents
            file_contents = file.read()
        finally:
            # closes the file
            file.close()

        # returns the file contents
        return file_contents

    def write_file(self, file_path, file_contents):
        """
        Writes the contents to the file in the
        given file path.

        :type file_path: String
        :param file_path: The path to the file to
        be written.
        :type file_contents: String
        :param file_contents: The contents to be
        written to the file.
        """

        # creates the directory for the file path
        self._create_directory(file_path)

        # open the file
        file = open(file_path, "wb")

        try:
            # writes the file contents
            file.write(file_contents)
        finally:
            # closes the file
            file.close()

    def remove_directory(self, directory_path, handle_exception = False):
        """
        Removes the directory in the given path.

        :type directory_path: String
        :param directory_path: The path to the directory
        to be removed.
        :type handle_exception: bool
        :param handle_exception: If an eventual raised exception shall
        be handled gracefully.
        """

        # in case the directory path does not exists
        if not os.path.exists(directory_path):
            # returns immediately
            return

        # retrieves the directory items
        directory_items = os.listdir(directory_path)

        try:
            # checks the directory for items and removes
            # the directory in the path (recursively)
            not directory_items and os.removedirs(directory_path)
        except Exception as exception:
            if not handle_exception: raise exception

    def remove_file(self, file_path, handle_exception = False):
        """
        Removes the file in the given path.

        :type directory_path: String
        :param directory_path: The path to the file
        to be removed.
        :type handle_exception: bool
        :param handle_exception: If an eventual raised exception shall
        be handled gracefully.
        """

        # in case the file path does not exists
        if not os.path.exists(file_path):
            # returns immediately
            return

        try:
            # removes the file path
            os.remove(file_path)
        except Exception as exception:
            if not handle_exception: raise exception

    def remove_directory_immediate(self, directory_path, handle_exception = False):
        """
        Removes the directory in the given directory path.

        :type directory_path: String
        :param directory_path: The path to the directory
        to be removed.
        :type handle_exception: bool
        :param handle_exception: If an eventual raised exception shall
        be handled gracefully.
        """

        # removes the directory in the (real) directory path
        path_util.remove_directory(directory_path, handle_exception)

    def get_file_path(self, file_path):
        """
        Retrieves a file path from the given file
        path.
        The returned file path is used in
        the current file context.

        :type file_path: String
        :param file_path: The file path to be used as base.
        :rtype: String
        :return: The file path to be used for writing.
        """

        # returns the file path
        return file_path

    def _create_directory(self, file_path):
        """
        Creates a directory from the given file path.

        :type file_path: String
        :param file_path: The path to the directory
        to be created.
        """

        # retrieves the directory path for the file path
        directory_path = os.path.dirname(file_path)

        # in case the directory path exists
        if os.path.exists(directory_path):
            # returns immediately
            return

        # creates the various required directories
        os.makedirs(directory_path)

class TransactionContext(object):
    """
    Generic transaction abstraction that hold the logic
    to the calling of the various transaction section
    callbacks for a "normal" workflow.
    """

    commit_callbacks_list = []
    """ The list of callbacks to be called upon commit """

    pre_commit_callbacks_list = []
    """ The list of callbacks to be called before commit """

    rollback_callbacks_list = []
    """ The list of callbacks to be called upon rollback """

    pre_rollback_callbacks_list = []
    """ The list of callbacks to be called before rollback """

    def __init__(self):
        """
        Constructor of the class.
        """

        self.commit_callbacks_list = []
        self.pre_commit_callbacks_list = []
        self.rollback_callbacks_list = []
        self.pre_rollback_callbacks_list = []

    def add_commit_callback(self, callback):
        """
        Adds a new commit callback.
        This callback will be called upon the final
        commit is passed.

        :type callback: Function
        :param callback: The callback function to be called
        upon the final commit.
        """

        self.remove_commit_callback(callback)
        self.commit_callbacks_list.append(callback)

    def remove_commit_callback(self, callback):
        """
        Removes an existing commit callback.

        :type callback: Function
        :param callback: The callback function to be called
        upon the final commit.
        """

        if not callback in self.commit_callbacks_list: return
        self.commit_callbacks_list.remove(callback)

    def add_pre_commit_callback(self, callback):
        """
        Adds a new pre commit callback.
        This callback will be called upon the final
        commit is started.

        :type callback: Function
        :param callback: The callback function to be called
        upon the final commit.
        """

        self.remove_pre_commit_callback(callback)
        self.pre_commit_callbacks_list.append(callback)

    def remove_pre_commit_callback(self, callback):
        """
        Removes an existing pre commit callback.

        :type callback: Function
        :param callback: The callback function to be called
        upon the final commit.
        """

        if not callback in self.pre_commit_callbacks_list: return
        self.pre_commit_callbacks_list.remove(callback)

    def add_rollback_callback(self, callback):
        """
        Adds a new rollback callback.
        This callback will be called upon the final
        rollback is passed.

        :type callback: Function
        :param callback: The callback function to be called
        upon the final rollback.
        """

        self.remove_rollback_callback(callback)
        self.rollback_callbacks_list.append(callback)

    def remove_rollback_callback(self, callback):
        """
        Removes an existing rollback callback.

        :type callback: Function
        :param callback: The callback function to be called
        upon the final rollback.
        """

        if not callback in self.rollback_callbacks_list: return
        self.rollback_callbacks_list.remove(callback)

    def add_pre_rollback_callback(self, callback):
        """
        Adds a new pre rollback callback.
        This callback will be called upon the final
        rollback is started.

        :type callback: Function
        :param callback: The callback function to be called
        upon the final rollback.
        """

        self.remove_pre_rollback_callback(callback)
        self.pre_rollback_callbacks_list.append(callback)

    def remove_pre_rollback_callback(self, callback):
        """
        Removes an existing pre rollback callback.

        :type callback: Function
        :param callback: The callback function to be called
        upon the final rollback.
        """

        if not callback in self.pre_rollback_callbacks_list: return
        self.pre_rollback_callbacks_list.remove(callback)

    def _call_commit_callbacks(self):
        """
        Calls all the commit callback functions
        in the current list.
        This method should be called at the
        end of a commit.
        """

        # iterates over all the commit callback functions
        for commit_callback in self.commit_callbacks_list:
            # calls the the commit callback
            commit_callback()

        # empties the commit callbacks
        self.commit_callbacks_list = []

    def _call_pre_commit_callbacks(self):
        """
        Calls all the pre commit callback functions
        in the current list.
        This method should be called at the
        beginning of a commit.
        """

        # iterates over all the pre commit callback functions
        for pre_commit_callback in self.pre_commit_callbacks_list:
            # calls the the pre commit callback
            pre_commit_callback()

        # empties the pre commit callbacks
        self.pre_commit_callbacks_list = []

    def _call_rollback_callbacks(self):
        """
        Calls all the rollback callback functions
        in the current list.
        This method should be called at the
        end of a rollback.
        """

        # iterates over all the rollback callback functions
        for rollback_callback in self.rollback_callbacks_list:
            # calls the the rollback callback
            rollback_callback()

        # empties the rollback callbacks
        self.rollback_callbacks_list = []

    def _call_pre_rollback_callbacks(self):
        """
        Calls all the pre rollback callback functions
        in the current list.
        This method should be called at the
        beginning of a rollback.
        """

        # iterates over all the pre rollback callback functions
        for pre_rollback_callback in self.pre_rollback_callbacks_list:
            # calls the the pre rollback callback
            pre_rollback_callback()

        # empties the pre rollback callbacks
        self.pre_rollback_callbacks_list = []

class FileImmediateContext(FileContext, TransactionContext):
    """
    The file immediate context class that controls
    an immediate transaction involving the file system.
    """

    def __init__(self):
        """
        Constructor of the class.
        """

        FileContext.__init__(self)
        TransactionContext.__init__(self)

    def open(self):
        """
        Opens a new transaction context.
        """

        pass

    def commit(self, remove_duplicates = True):
        """
        Commits a new transaction context.
        All the pending file operations
        will be persisted.

        An optional argument may be unset if
        the duplicate operation are not meant
        to be removed.

        :type remove_duplicates: bool
        :param remove_duplicats: If duplicate operations
        are meant to be removed so that no duplicate
        operation are performed (performance impact).
        """

        # calls the pre (before) commit callbacks
        self._call_pre_commit_callbacks()

        # calls the "final" commit callbacks
        self._call_commit_callbacks()

    def rollback(self):
        """
        Reverts all the pending operations in
        the current transaction context.
        All the data pending persistence will
        be dropped.
        """

        # calls the pre (before) rollback callbacks
        self._call_pre_rollback_callbacks()

        # calls the "final" rollback callbacks
        self._call_rollback_callbacks()

class FileTransactionContext(FileContext, TransactionContext):
    """
    The file transaction context class that controls
    a transaction involving the file system.
    """

    transaction_level = 0
    """ The current transaction level in use """

    temporary_path = None
    """ The temporary path to be used in the file transaction """

    path_tuples_list = []
    """ The list of path tuples associated with the transaction """

    access_lock = None
    """ The lock controlling the access to the file transaction """

    def __init__(self, temporary_path = None):
        """
        Constructor of the class.

        :type temporary_path: String
        :param temporary_path: The temporary path to be used
        for the transaction temporary files.
        """

        FileContext.__init__(self)
        TransactionContext.__init__(self)
        self.temporary_path = temporary_path or tempfile.mkdtemp()

        self.path_tuples_list = []
        self.access_lock = threading.RLock()

    def resolve_file_path(self, file_path):
        """
        Resolves the given file path, checking if it
        exists and if it is valid.

        :type file_path: String
        :param file_path: The file path to be resolved.
        :rtype: String
        :return: The resolved (real) file path.
        """

        # retrieves the virtual file path for the file path
        virtual_file_path = self._get_virtual_file_path(file_path)

        # checks if the virtual file path exists
        virtual_file_path_exists = os.path.exists(virtual_file_path)

        # resolves the file path taking into account the
        # existence of the virtual file path
        real_file_path = virtual_file_path_exists and virtual_file_path or file_path

        # returns the real file path
        return real_file_path

    def exists_file_path(self, file_path):
        """
        Tests if the given file path exists in
        the current virtual environment.

        :type file_path: String
        :param file_path: The file path to be tested
        for existence.
        :rtype: bool
        :return: The result of the test for
        file existence.
        """

        # resolves the virtual file path
        virtual_file_path = self._get_virtual_file_path(file_path)

        # tests if the file has been already (potentially) removed
        removed_file_path = self._is_removed_file_path(file_path)

        # tests if both the (real) file path and the virtual
        # file path exist
        file_path_exists = os.path.exists(file_path)
        virtual_file_path_exists = os.path.exists(virtual_file_path)

        # tests if the file path exists, checks the condition
        # that if the (real) file path exists the removed file path
        # must not exist otherwise if the virtual file path exists
        # the file also exists
        exists_file_path = (file_path_exists and not removed_file_path) or virtual_file_path_exists

        # returns the exists file path result
        return exists_file_path

    def is_directory_path(self, file_path):
        """
        Tests if the given file path refers a directory path in
        the current virtual environment.

        :type file_path: String
        :param file_path: The file path to be tested
        for directory referral.
        :rtype: bool
        :return: The result of the test for
        directory referral.
        """

        # resolves the file path (real file path)
        real_file_path = self.resolve_file_path(file_path)

        # tests if the file path refers a directory path
        is_directory_path = os.path.isdir(real_file_path)

        # returns the is directory path result
        return is_directory_path

    def read_file(self, file_path):
        """
        Reads the given file contents from a file.
        In case a transaction exists the contents read
        are the virtual ones.

        :type file_path: String
        :param file_path: The path to the file to be
        read.
        :rtype: String
        :return: The read file contents.
        """

        # resolves the file path (real file path)
        real_file_path = self.resolve_file_path(file_path)

        # reads the file using the file context (virtual file path used)
        file_contents = FileContext.read_file(self, real_file_path)

        # returns the file contents
        return file_contents

    def write_file(self, file_path, file_contents, replace_file = True):
        """
        Writes the given file contents to a file in
        the given file path.
        This write is not persisted immediately and
        pushed to a transaction.

        :type file_path: String
        :param file_path: The path to the file to be
        written.
        :type file_contents: String
        :param file_contents: The contents to be written
        to the file.
        :type replace_file: bool
        :param replace_file: If the file should be replaced
        in case existent file is found.
        """

        # retrieves the virtual file path for the file path
        virtual_file_path = self._get_virtual_file_path(file_path)

        # writes the file using the file context (virtual file path used)
        FileContext.write_file(self, virtual_file_path, file_contents)

        # sets the operation based on the value of the replace file flag
        operation = replace_file and ADD_OPERATION or ADD_NO_REPLACE_OPERATION

        # creates a path tuple with the virtual file path
        # and the file path for the operation add
        path_tuple = (
            operation,
            virtual_file_path,
            file_path
        )

        # adds the path tuple
        self._add_path_tuple(path_tuple)

    def remove_directory(self, directory_path, handle_exception = False):
        """
        Removes the directory in the given path.
        This removal is not persisted immediately and
        pushed to a transaction.

        :type directory_path: String
        :param directory_path: The path to the directory
        to be removed.
        :type handle_exception: bool
        :param handle_exception: If an eventual raised exception shall
        be handled gracefully.
        """

        # retrieves the virtual directory path for the file path
        virtual_directory_path = self._get_virtual_file_path(directory_path)

        # in case the virtual directory path exists
        if os.path.exists(virtual_directory_path):
            # retrieves the (virtual) directory items
            virtual_directory_items = os.listdir(virtual_directory_path)

            try:
                # checks the virtual directory for items and removes
                # the directory in the virtual path (recursively)
                not virtual_directory_items and os.removedirs(virtual_directory_path)
            except Exception as exception:
                if not handle_exception: raise exception

        # creates a path tuple with the directory path
        # the operation remove, the recursive flag is set
        path_tuple = (
            REMOVE_OPERATION,
            directory_path,
            handle_exception,
            True
        )

        # adds the path tuple
        self._add_path_tuple(path_tuple)

    def remove_file(self, file_path, handle_exception = False):
        """
        Removes the file in the given path.
        This removal is not persisted immediately and
        pushed to a transaction.

        :type directory_path: String
        :param directory_path: The path to the file
        to be removed.
        :type handle_exception: bool
        :param handle_exception: If an eventual raised exception shall
        be handled gracefully.
        """

        # retrieves the virtual file path for the file path
        virtual_file_path = self._get_virtual_file_path(file_path)

        # in case the virtual file path exists
        if os.path.exists(virtual_file_path):
            try:
                # removes the virtual file path
                os.remove(virtual_file_path)
            except Exception as exception:
                if not handle_exception: raise exception

        # creates a path tuple with the file path
        # the operation remove, the recursive flag is unset
        path_tuple = (
            REMOVE_OPERATION,
            file_path,
            handle_exception,
            False
        )

        # adds the path tuple
        self._add_path_tuple(path_tuple)

    def remove_directory_immediate(self, directory_path, handle_exception = False):
        """
        Removes the directory in the given directory path.
        In case a transaction exists the directory to be
        removed is the virtual one.

        :type directory_path: String
        :param directory_path: The path to the directory
        to be removed.
        :type handle_exception: bool
        :param handle_exception: If an eventual raised exception shall
        be handled gracefully.
        """

        # resolves the directory path (real directory path)
        real_directory_path = self.resolve_file_path(directory_path)

        try:
            # removes the directory in the (real) directory path
            path_util.remove_directory(real_directory_path)
        except Exception as exception:
            if not handle_exception: raise exception

    def get_file_path(self, file_path, replace_files = True):
        """
        Retrieves a file path to be used for writing.
        The file path to be used will be used in a
        transaction environment and context.

        :type file_path: String
        :param file_path: The file path to be used as base.
        :type replace_files: bool
        :param replace_files: If the files should be replaced
        in case duplicate files are found.
        :rtype: String
        :return: The file path to be used for writing.
        """

        # retrieves the virtual file path for the file path
        virtual_file_path = self._get_virtual_file_path(file_path)

        # sets the operation based on the value of the replace files flag
        operation = replace_files and ADD_OPERATION or ADD_NO_REPLACE_OPERATION

        # creates a path tuple with the virtual file path
        # and the file path
        path_tuple = (
            operation,
            virtual_file_path,
            file_path
        )

        # adds the path tuple
        self._add_path_tuple(path_tuple)

        # returns the virtual file path
        return virtual_file_path

    def open(self):
        """
        Opens a new transaction context.
        """

        # acquires the access lock
        self.access_lock.acquire()

        try:
            # increments the transaction level
            self.transaction_level += 1
        finally:
            # releases the access lock
            self.access_lock.release()

    def commit(self, remove_duplicates = True):
        """
        Commits a new transaction context.
        All the pending file operations
        will be persisted.

        An optional argument may be unset if
        the duplicate operation are not meant
        to be removed.

        :type remove_duplicates: bool
        :param remove_duplicats: If duplicate operations
        are meant to be removed so that no duplicate
        operation are performed (performance impact).
        """

        # acquires the access lock
        self.access_lock.acquire()

        try:
            # decrements the transaction level
            self.transaction_level -= 1

            # in case the transaction level is positive
            if self.transaction_level > 0:
                # returns immediately
                return
            # in case the transaction level is negative
            elif self.transaction_level < 0:
                # raises the runtime error
                raise RuntimeError("Invalid transaction level")

            # calls the pre (before) commit callbacks
            self._call_pre_commit_callbacks()

            # creates a no duplicates list from the path tuples
            # list, this will ensure that no duplicate operations
            # exist (this is a critical performance trick)
            path_tuples_list = remove_duplicates and list_util.list_no_duplicates(self.path_tuples_list) or self.path_tuples_list

            # iterates over all the path tuples in
            # path tuples list
            for path_tuple in path_tuples_list:
                # retrieves the operation (name)
                operation = path_tuple[0]

                # creates the path tuple process method name from the operation
                path_tuple_process_method_name = PATH_TUPLE_PROCESS_METHOD_PREFIX + operation

                # retrieves the path tuple process method
                path_tuple_process_method = getattr(self, path_tuple_process_method_name)

                # calls the path tuple process method
                path_tuple_process_method(path_tuple)

            # runs the cleanup
            self._cleanup()

            # runs the reset
            self._reset()

            # calls the "final" commit callbacks
            self._call_commit_callbacks()
        finally:
            # releases the access lock
            self.access_lock.release()

    def rollback(self):
        """
        Reverts all the pending operations in
        the current transaction context.
        All the data pending persistence will
        be dropped.
        """

        # acquires the access lock
        self.access_lock.acquire()

        try:
            # in case the transaction level is zero
            if self.transaction_level == 0:
                # returns immediately
                return
            # in case the transaction level is negative
            elif self.transaction_level < 0:
                # raises the runtime error
                raise RuntimeError("Invalid transaction level")

            # calls the pre (before) rollback callbacks
            self._call_pre_rollback_callbacks()

            # runs the cleanup
            self._cleanup()

            # runs the reset
            self._reset()

            # calls the "final" rollback callbacks
            self._call_rollback_callbacks()
        finally:
            # releases the access lock
            self.access_lock.release()

    def _reset(self):
        """
        Resets the state of the transaction file.
        """

        # empties the path tuples list
        self.path_tuples_list = []

        # resets the transaction level
        self.transaction_level = 0

        # in case the temporary path is a directory removes
        # the temporary path
        os.path.isdir(self.temporary_path) and path_util.remove_directory(self.temporary_path)

    def _cleanup(self):
        """
        Cleans the current temporary path.
        All the data in the temporary path
        will be deleted.
        """

        # retrieves the temporary path items
        temporary_path_items = os.listdir(self.temporary_path)

        # iterates over all the temporary path items
        for temporary_path_item in temporary_path_items:
            # creates the temporary complete path item, by joining the
            # temporary path and the temporary path item
            temporary_complete_path_item = os.path.join(self.temporary_path, temporary_path_item)

            # in case the path does not exist (no need to proceed
            # with removal)
            if not os.path.exists(temporary_complete_path_item):
                # continues the loop
                continue

            # in case the temporary item is a directory
            if os.path.isdir(temporary_complete_path_item):
                # removes the directory in the temporary
                # complete path item
                path_util.remove_directory(temporary_complete_path_item)
            # otherwise it must be a "normal" file
            else:
                # removes the temporary complete path item
                os.remove(temporary_complete_path_item)

    def _add_path_tuple(self, path_tuple):
        """
        Adds a path tuple to the current path
        tuples list.

        :type path_tuple: Tuple
        :param path_tuple: The path tuple to
        be added to the tuples list.
        """

        # adds the path tuple to the path tuples list
        self.path_tuples_list.append(path_tuple)

    def _get_virtual_file_path(self, file_path):
        """
        Retrieves the "virtual" file path for the given
        file path.

        :type file_path: String
        :param file_path: The file path to retrieve
        the "virtual" file path.
        :rtype: String
        :return: The virtual file path.
        """

        # splits the file path into drive and
        # base file path
        _drive, base_file_path = os.path.splitdrive(file_path)

        # strips the base file path
        base_file_path = base_file_path.lstrip("\\/")

        # joins the temporary path and the
        # base file path
        virtual_file_path = os.path.join(self.temporary_path, base_file_path)

        # normalizes the virtual file path
        virtual_file_path = os.path.normpath(virtual_file_path)

        # returns the virtual file path
        return virtual_file_path

    def _is_removed_file_path(self, file_path):
        """
        Checks if the given file path has been removed in
        the current context.
        The fact that the file has been removed in the
        current context foes not imply that it does not exists
        in the context.
        If a write operation has been made after a removal
        the file should be accounted as existent.

        This is a very slow operation and should be used carefully
        otherwise the operation may become clogged.

        :type file_path: String
        :param file_path: The path to the file to be checked
        for removal.
        :rtype: bool
        :return: The result of the file removal check.
        """

        # iterates over all the path tuples in
        # path tuples list to check for the ones
        # that may remove the path
        for path_tuple in self.path_tuples_list:
            # retrieves the first item from the path
            # tuple, the operation value
            operation = path_tuple[0]

            # in case the operation is not of type remove
            # continues the loop no removal
            if not operation == REMOVE_OPERATION: continue

            # unpacks the path tuple (for a remove operation)
            operation, _file_path, _handle_exception, recursive = path_tuple

            # in case the recursive flag is not set (possibly just a
            # normal file an not a directory) and the file path does
            # not matches the one required skips extra verification
            if not recursive and not file_path == _file_path: continue

            # checks if the current iteration file path is
            # a parent path to the path being checked
            is_parent_path = path_util.is_parent_path(file_path, _file_path)

            # in case it's not a parent path continues
            # the loop no removal detected
            if not is_parent_path: continue

            # removal detected
            return True

        # returns false (no removal detected)
        return False

    def _process_path_tuple_add(self, path_tuple):
        """
        Processes a path tuple of type add.
        The remote tuple refers paths to be added
        in persistence.

        :type path_tuple: String
        :param path_tuple: The tuple with the path contents.
        """

        # unpacks the path tuple
        _operation, virtual_file_path, file_path = path_tuple

        # in case the path does not exist (no need to proceed
        # with persistence), it represents a case where a temporary
        # path was request but not used for writing
        if not os.path.exists(virtual_file_path):
            # returns immediately
            return

        # in case the virtual file path is a directory
        if os.path.isdir(virtual_file_path):
            # copies the directory in the virtual path to the directory in the file path
            path_util.copy_directory(virtual_file_path, file_path)
        # otherwise it must be a "normal" file
        else:
            # copies the file in the virtual path to the file in the file path
            path_util.copy_file(virtual_file_path, file_path)

    def _process_path_tuple_add_no_replace(self, path_tuple):
        """
        Processes a path tuple of type add no replace.
        The remote tuple refers paths to be added
        in persistence.
        In this mode no files are replaced in case of name
        collision.

        :type path_tuple: String
        :param path_tuple: The tuple with the path contents.
        """

        # unpacks the path tuple
        _operation, virtual_file_path, file_path = path_tuple

        # in case the path does not exist (no need to proceed
        # with persistence), it represents a case where a temporary
        # path was request but not used for writing
        if not os.path.exists(virtual_file_path):
            # returns immediately
            return

        # in case the virtual file path is a directory
        if os.path.isdir(virtual_file_path):
            # copies the directory in the virtual path to the directory in the file path
            # the copy of the directory is made without replacing existent files
            path_util.copy_directory(virtual_file_path, file_path, False)
        # otherwise it must be a "normal" file
        else:
            # copies the file in the virtual path to the file in the file path
            # the copy of the file is made without replacing an eventual
            # existent file
            path_util.copy_file(virtual_file_path, file_path, False)

    def _process_path_tuple_remove(self, path_tuple):
        """
        Processes a path tuple of type remove.
        The remote tuple refers paths to be removed
        in persistence.

        :type path_tuple: String
        :param path_tuple: The tuple with the path contents.
        """

        # unpacks the path tuple
        _operation, file_path, handle_exception, recursive = path_tuple

        # in case the path does not exist (no need to proceed
        # with removal)
        if not os.path.exists(file_path):
            # returns immediately
            return

        # in case the file path is a directory
        if os.path.isdir(file_path):
            # retrieves the directory items
            directory_items = os.listdir(file_path)

            # in case the recursive mode is active
            if recursive:
                try:
                    # checks the directory for items and removes
                    # the directory in the path (recursively)
                    not directory_items and os.removedirs(file_path)
                except Exception as exception:
                    if not handle_exception: raise exception
            # in case the removal is not recursive
            else:
                try:
                    # checks the directory for items and removes
                    # the directory in the path
                    not directory_items and os.remove(file_path)
                except Exception as exception:
                    if not handle_exception: raise exception
        # otherwise it must be a "normal" file
        else:
            try:
                # removes the file path
                os.remove(file_path)
            except Exception as exception:
                if not handle_exception: raise exception



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import time
import threading

from . import verify_util

SCHEDULING_MAX = getattr(threading, "TIMEOUT_MAX", 3600)
""" The value for the maximum timeout value allowed
for the threading await operations (defaults to an hour
in case no base value is obtainable) """

class Scheduler(threading.Thread):
    """
    Class that implements a scheduler to be used
    to "call" callable objects for a provided timestamp.
    """

    running_flag = False
    """ Flag that controls if the scheduler event loop is currently
    running, set at the start of the loop and unset by the end of it """

    continue_flag = False
    """ Flag controlling the execution of the scheduler, if unset
    it will trigger the unloading of the loop """

    busy_flag = False
    """ Flag controlling the busy state of the scheduler, controls
    if there's working being "done" by the scheduler """

    timestamp_queue = []
    """ Ordered list (queue) of timestamps for callables """

    timestamp_map = {}
    """ The map associating the timestamp with a list of callables """

    tasks = set()
    """ The complete set of tasks that are currently under pending
    execution, this should be a set of unique identifiers """

    waits = set()
    """ The sequence that contains the complete set of callables that
    are waiting to be notified """

    condition = None
    """ The condition that will control the access to the data structures
    and trigger events on the production of new items """

    waits_condition = None
    """ Condition that is going to be used in the waits operation and control
    the access to the waits set """

    exception_handler = None
    """ If set defined an handler (callable) that is going to be called
    whenever an exception is raised in the execution of the callable units """

    _counter = 1
    """ The unique identifier counter that is going to be incremented
    per each callable to be added """

    def __init__(self):
        """
        Constructor of the class.

        :type plugin: Plugin
        :param plugin: The plugin to be used.
        """

        threading.Thread.__init__(self)

        self.daemon = True
        self.timestamp_queue = []
        self.timestamp_map = {}
        self.tasks = set()
        self.waits = set()
        self.condition = threading.Condition()
        self.waits_condition = threading.Condition()

    def run(self):
        # sets the initial value of the timeout, which is an unset
        # one, this value is only going to be set in case of deferred
        # first item in queue exists
        timeout = None

        # sets the running flag, effectively indicating that the scheduler
        # is running its main loop
        self.running_flag = True

        try:
            # iterates while the continue flag is set, this means
            # that this is a continuous loop operation
            while self.continue_flag:

                # acquires the condition so that we can safely wait for
                # new "events" and access the underlying data structures
                # for proper and safe consuming of them
                with self.condition:
                    # waits for the condition while the system is running and
                    # either the queue is empty or there's a timeout defined
                    # (meaning a pending final value has been reached)
                    while self.continue_flag and (not self.timestamp_queue or timeout):
                        # makes sure that the timeout value does not overflow
                        # maximum timeout value for scheduling, this would
                        # raise an overflow exception, making this verification
                        # will imply running more wait operations for large values
                        # which is OK as no significant resources will be used
                        if timeout: timeout = min(SCHEDULING_MAX, timeout)

                        # waits for the condition and timeouts according to the
                        # provided partial value (if any)
                        self.condition.wait(timeout = timeout)

                        # resets the timeout value as we've stepped outside the
                        # wait operation and this value is no longer relevant
                        timeout = None

                    # in case the continue flag has been unset
                    # (by triggering condition), then breaks loop
                    if not self.continue_flag: break

                    # retrieves the current timestamp, to be
                    # used in comparison operations
                    current_timestamp = time.time()

                    # retrieves the timestamp from the
                    # timestamp queue
                    timestamp = self.timestamp_queue[0]

                    # in case the final timestamp has been
                    # reached, meaning that the timestamp
                    # of the task is greater than the current
                    # timestamp
                    if timestamp > current_timestamp:
                        # sleeps for the amount of time defined
                        # in the sleep step
                        timeout = timestamp - current_timestamp

                        # breaks the loop (no more work
                        # to be processed for now)
                        continue

                    # pops (removes first element) the timestamp
                    # from the timestamp queue, proper timestamp
                    # consuming operation
                    self.timestamp_queue.pop(0)

                    # retrieves the callable (elements) list
                    # for the timestamp
                    callable_list = self.timestamp_map[timestamp]

                    # removes the callable list for the timestamp
                    # (done before the calling to avoid race condition)
                    del self.timestamp_map[timestamp]

                # runs the callable calling operation (consuming)
                # which should properly handle exceptions avoiding
                # internal execution problems
                self._handle_callables(callable_list)
        finally:
            self.running_flag = False
            self.continue_flag = False

    def start_scheduler(self):
        """
        Starts the scheduler process.
        This method is asynchronous and the starting of
        the scheduler is not immediate.
        This method creates a new thread for scheduling.
        """

        # if the scheduler is already running avoids
        # duplicate starting, returns immediately
        if self.continue_flag: return

        # sets the continue flag to the valid value (controls
        # the loop) and then start the main loop process
        self.continue_flag = True
        self.start()

    def stop_scheduler(self):
        """
        Stops the scheduler process.
        This method is asynchronous and the stopping of
        the scheduler is not immediate.
        """

        # if the scheduler is already stopped avoids
        # duplicate stopping, returns immediately
        if not self.continue_flag: return

        # unsets the continue flag, this will trigger
        # the unload of the scheduler as soon as possible
        self.continue_flag = False

        # triggers the condition so that the listener is
        # able to stop the waiting process
        with self.condition:
            self.condition.notify()

    def reset_scheduler(self, notify = True):
        """
        Resets the scheduler to the original state.
        This method may be used to avoid the allocation
        of new scheduler objects.
        This operation is dangerous an is not thread safe.
        A typical usage of this method involves first the
        stopping of the scheduler.

        :type notify: bool
        :param notify: If a notification in the condition should
        be sent so that the thread is unblocked if needed.
        """

        _condition = self.condition

        self.running_flag = False
        self.continue_flag = False
        self.busy_flag = False
        self.timestamp_queue = []
        self.timestamp_map = {}
        self.tasks = set()
        self.waits = set()
        self.condition = threading.Condition()
        self.wait_execution = threading.Condition()
        self.exception_handler = None
        self._counter = 1

        if notify:
            with _condition: _condition.notify()

    def add_callable(self, callable, timestamp = None, verify = True):
        """
        Adds a callable object to the scheduler
        for calling upon the given timestamp value.
        The sent callable is called without any arguments
        and the real time for calling may not be assured.

        An unique callable identifier is returned.

        :type callable: Callable
        :param callable: The callable object to be called
        upon in time described in the given timestamp.
        :type timestamp: float
        :param timestamp: The timestamp describing the
        time for calling the callable object, if not defined
        the current timestamp is used (immediate scheduling).
        :type ensure: bool
        :param ensure: If set makes sure that the scheduler
        is running, raising an exception otherwise.
        :rtype: int
        :return: The identifier of the callable task that has just
        been created, can be used for waiting for task completion.
        """

        # in case the verify flag is set then we need to make
        # sure that the scheduler is running
        if verify: verify_util.verify(self.is_running())

        # in case no explicit timestamp is provided then uses
        # the current time - immediate task scheduling
        if timestamp == None: timestamp = time.time()

        # acquires the condition to be able to safely
        # manipulate the structure and produce item
        with self.condition:
            # obtains the identifier for the current
            # callable operation schedule
            identifier = self._counter
            self._counter += 1

            # starts the index value
            index = 0

            # iterates over all the timestamps in the timestamp
            # queue (to find position for insertion)
            for _timestamp in self.timestamp_queue:
                # in case the the current iteration timestamp
                # contains a value smaller than the timestamp
                # to be inserted
                if timestamp < _timestamp:
                    # breaks the loop (position as
                    # insertion has been reached)
                    break

                # increments the index, offsetting the value
                index += 1

            # checks if the timestamp already exists in the
            # current structures
            timestamp_exists = timestamp in self.timestamp_map

            # inserts the timestamp in the timestamp queue
            # for the correct index (in order to maintain order)
            # in case it does not exist already
            if not timestamp_exists: self.timestamp_queue.insert(index, timestamp)

            # retrieves the list of callable for the given timestamp
            # and then updates it with the given callable object
            callable_list = self.timestamp_map.get(timestamp, [])
            callable_list.append((callable, identifier))
            self.timestamp_map[timestamp] = callable_list

            # adds the identifier to the sequence that controls the
            # tasks that are considered active
            self.tasks.add(identifier)

            # notifies the condition effectively indicating that a
            # new item or set of items is available for consumption
            self.condition.notify()

            # returns the final identifier for the callable task
            # that has just been scheduled
            return identifier

    def wait_callable(self, identifier):
        verify_util.verify(isinstance(identifier, int))

        with self.waits_condition:
            self.waits.add(identifier)

        if not identifier in self.tasks:
            with self.waits_condition:
                if identifier in self.waits:
                    self.waits.remove(identifier)
            return

        with self.waits_condition:
            while self.continue_flag and identifier in self.waits:
                self.waits_condition.wait()

    def set_exception_handler(self, exception_handler):
        """
        Sets the handler that will be called in case there's an
        user exception raised in the callable.

        :type exception_handler: Callable
        :param exception_handler: The handler to be called upon
        exception in callable execution.
        """

        self.exception_handler = exception_handler

    def is_busy(self):
        """
        Checks if the scheduler is currently in a busy status
        meaning that it's executing some sort of work.

        :rtype: bool
        :return: If the scheduler is executing any kind of work.
        """

        return self.busy_flag

    def is_running(self, pedantic = False):
        """
        Checks if the scheduler is currently running, the scheduler
        is considered to be running if the continue flag is set.

        In case the pedantic flag is set the scheduler is only
        considered to be running in case the running flag (which
        is set when thread is running) is also set.

        :rtype: bool
        :return: If the scheduler is currently running.
        """

        if pedantic: return self.running_flag and self.continue_flag
        return self.continue_flag

    def _handle_callables(self, callable_list):
        # sets the busy flag indicating that there's execution
        # of callable object happening
        self.busy_flag = True

        try:
            # iterates over all the callables to call
            # them (calls the proper function)
            for callable, _identifier in callable_list:
                try:
                    # calls the callable (element)
                    # this can be of long duration
                    callable()
                except Exception as exception:
                    if self.exception_handler:
                        self.exception_handler(callable, exception)
                    else:
                        print(exception)
        finally:
            self.busy_flag = False

        # runs a final waits condition operation that will
        # make sure that the pending waits values are notified
        # in case they are in a waiting state, it will also
        # remove the multiple callable tasks from the sequence
        # that controls the pending tasks
        with self.waits_condition:
            notify = False

            for callable, identifier in callable_list:
                if identifier in self.waits:
                    self.waits.remove(identifier)
                    notify = True
                self.tasks.remove(identifier)

            if notify:
                self.waits_condition.notify_all()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import sys

from colony.base import legacy

def __import__(module_name, persist_value = True):
    """
    Importer function to be used in the process of importing
    a module referred in inverted way.
    The optional persist value may be used to control if the
    globals/locals reference value must be set in the caller module
    in case it has been retrieved from a parent caller (cache).
    This function should be used in cases where the inversion injection
    was made using the data helper.

    :type module_name: String
    :param module_name: The name of the module to be imported.
    :type persist_value: bool
    :param persist_value: If the globals/locals value shall be
    persisted in the caller in case it's is not available there.
    :rtype: module
    :return: The imported module.
    """

    # unsets the module reference
    module = None

    # starts the index counter value
    # to start in the previous caller
    index = 1

    try:
        # iterates continuously over the stack frame to try to gather
        # any loading frame that refers the request module name in it's
        # global or locals names structure (reversed importing process)
        while True:
            # retrieves the caller of the importer method
            caller = sys._getframe(index)

            # in case the module name exists in the globals map
            # of the caller
            if module_name in caller.f_globals.get("_globals", {}):
                # retrieves the module from the globals map of the caller and
                # then breaks the current loop to return the module reference
                module = caller.f_globals["_globals"][module_name]
                break

            # in case the module name exists in the locals map
            # of the caller
            elif module_name in caller.f_globals.get("_locals", {}):
                # retrieves the module from the locals map of the caller and
                # then breaks the current loop to return the module reference
                module = caller.f_globals["_locals"][module_name]
                break

            # increments the index counter so that the stack position
            # is incremented by one more value (one more level)
            index += 1
    except ValueError:
        # raises a runtime error because it could
        # not retrieve the module
        raise ImportError("No module named '%s' found in global or local references" % module_name)

    # raise a runtime error in case the module
    # is not found (problem in the import)
    if not module: raise ImportError("No module named '%s' found in global or local references" % module_name)

    # in case the module value was retrieved from an upper
    # calling layer and the persist value flag is set (cache)
    # the module value must be persisted in the direct caller
    if index > 1 and persist_value:
        # retrieves the (direct) caller of the importer method
        # and sets the module in the globals reference value
        caller = sys._getframe(1)
        globals_reference = caller.f_globals.get("_globals", {})
        globals_reference[module_name] = module
        caller.f_globals["_globals"] = globals_reference

    # returns the module to the caller method that requested
    # the importing of the module (as expected)
    return module

def reload_import(path, hard = True):
    """
    Reloads the import (module) referred in the system modules
    by the given path.
    This reload is triggered synchronously and so at the return
    of this function the module should be reloaded.

    There are two modes of operation for reloading:
    The "hard" operation mode removes the module from the system
    modules and updates the environment.
    The "soft" operation modes "just" reloads the module in the
    current environment.

    :type path: String
    :param path: The (package) path to the module to be reloaded.
    This path should be the same as the one referring the module
    in the system modules.
    :type hard: bool
    :param hard: If an "hard" reload approach must be taken to provide
    the reload process. The hard reload approach removes the path
    from the system modules map.
    """

    # in case the path is not present in the
    # system modules no need to reload
    if not path in sys.modules: return

    # in case the hard approach for reloading is
    # taken the system modules should be changed
    if hard:
        # retrieves the module for the given path from
        # system module and then removes it from the system
        # modules and then deletes it from the virtual
        # machine environment
        module = sys.modules[path]
        del sys.modules[path]
        del module
    # otherwise the "soft" reload provides the normal
    # module reload method
    else:
        # retrieves the module for the given path from
        # system module and then forces a reload on the
        # module (to flush the contents)
        module = sys.modules[path]
        legacy.reload(module)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Hive Colony Framework
# Copyright (c) 2008-2020 Hive Solutions Lda.
#
# This file is part of Hive Colony Framework
#
# Hive Colony Framework is free software: you can redistribute it and/or modify
# it under the terms of the Apache License as published by the Apache
# Foundation, either version 2.0 of the License, or (at your option) any
# later version.
#
# Hive Colony Framework is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# Apache License for more details.
#
# You should have received a copy of the Apache License along with
# Hive Colony Framework If not, see <http://www.apache.org/licenses/>.

__author__ = "João Magalhães <joamag@hive.pt>"
""" The author(s) of the module """

__version__ = "1.0.0"
""" The version of the module """

__revision__ = "$LastChangedRevision$"
""" The revision number of the module """

__date__ = "$LastChangedDate$"
""" The last change date of the module """

__copyright__ = "Copyright (c) 2008-2020 Hive Solutions Lda."
""" The copyright for the module """

__license__ = "Apache License, Version 2.0"
""" The license for the module """

import re
import hashlib

from colony.base import legacy

HASH_VALUE = "hash"
""" The hash value """

VALUE_VALUE = "value"
""" The value value """

PLAIN_VALUE = "plain"
""" The plain value """

MD5_VALUE = "md5"
""" The MD5 value """

SHA1_VALUE = "sha1"
""" The SHA1 value """

SHA256_VALUE = "sha256"
""" The SHA256 value """

MD5_CRPYT_SEPARATOR = "$"
""" The MD5 crypt separator """

DEFAULT_MD5_CRYPT_MAGIC = "$1$"
""" The default MD5 crypt magic value """

DEFAULT_HASH_SET = (MD5_VALUE, SHA1_VALUE, SHA256_VALUE)
""" The default hash set """

INTEGER_TO_ASCII_64 = "./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
""" The array of conversion from integer to ascii """

PASSWORD_VALUE_REGEX_VALUE = "^\{(?P<hash>\w+)\}(?P<value>.+)$"
""" The password value regex value """

NUMBER_REGEX_VALUE = "\d+"
""" The number regex value """

LETTER_LOWER_REGEX_VALUE = "[a-z]"
""" The letter lower regex value """

LETTER_UPPER_REGEX_VALUE = "[A-Z]"
""" The letter upper regex value """

SPECIAL_CHARACTER_REGEX_VALUE = ".[!,@,#,$,%,^,&,*,?,_,~,-,£,(,)]"
""" The special character regex value """

PASSWORD_VALUE_REGEX = re.compile(PASSWORD_VALUE_REGEX_VALUE)
""" The password value regex """

NUMBER_REGEX = re.compile(NUMBER_REGEX_VALUE)
""" The number regex """

LETTER_LOWER_REGEX = re.compile(LETTER_LOWER_REGEX_VALUE)
""" The letter lower regex """

LETTER_UPPER_REGEX = re.compile(LETTER_UPPER_REGEX_VALUE)
""" The letter upper regex """

SPECIAL_CHARACTER_REGEX = re.compile(SPECIAL_CHARACTER_REGEX_VALUE)
""" The special character regex """

def password_crypt(password, salt = "", hash_method = MD5_VALUE):
    """
    Encrypts the given password using the provided hash method.
    An optional salt may be provided for extra security.
    The generated hash is always defined in hexadecimal.

    :type password: String
    :param password: The password to be encrypted using
    the hash method.
    :type salt: String
    :param salt: The salt to be used during the encryption
    process.
    :type hash_method: String
    :param hash_method: The name of the hash method to be used
    for encryption.
    :rtype: String
    :return: The generated (complete) hash hexadecimal string.
    """

    # converts the name of the hash method to lower
    # case string
    hash_method_lower = hash_method.lower()

    # creates the password (word) from the
    # password an the salt
    password_word = password + salt

    # in case the hash method is of type plain
    if hash_method_lower == PLAIN_VALUE:
        # sets the hash value as the (base)
        # password word value (plain)
        hash_value = password_word
    # otherwise it must be a general hash function
    else:
        # creates the new hash object from the
        # hash method
        hash = hashlib.new(hash_method)

        # converts the password word into a bytes
        # based string (if required) and updates
        # the hash value with the password word
        password_word = legacy.bytes(password_word)
        hash.update(password_word)

        # retrieves the hash value from the
        # hex digest
        hash_value = hash.hexdigest()

    # creates the final password hash prepending the
    # hash method reference
    password_hash = "{" + hash_method_lower + "}" + hash_value

    # returns the password (final) hash
    # value (with the hash method prefix)
    return password_hash

def password_match(password_hash, password, salt = ""):
    """
    Checks if the given password hash value matched
    the given password using the given (optional) salt.
    The matching process executes the original hash function
    in order to check for same results.

    :type password_hash: String
    :param password_hash: The complete password hash hexadecimal string.
    :type password: String
    :param password: The base password for checking.
    :type salt: String
    :param salt: The base salt for checking.
    :rtype: bool
    :return: The result of the password match checking.
    """

    # tries to match the base password hash
    base_password_match = PASSWORD_VALUE_REGEX.match(password_hash)

    # retrieves the base password hash and value
    base_password_hash = base_password_match.group(HASH_VALUE)
    base_password_value = base_password_match.group(VALUE_VALUE)

    # creates the password (word) from the
    # password an the salt (secure work)
    password_word = password + salt

    # sets the initial value for the passwords
    # math result
    passwords_match = False

    # in case the base password hash is of type plain
    if base_password_hash == PLAIN_VALUE:
        # checks if both passwords match
        passwords_match = password_word == base_password_value
    # otherwise it must be a general hash function
    else:
        # creates the new hash object from the
        # base password hash (method)
        hash = hashlib.new(base_password_hash)

        # updates the hash value with the
        # password word, note that the value
        # is ensured to be a valid byte value
        password_word = legacy.bytes(password_word)
        hash.update(password_word)

        # retrieves the hash value from the
        # hex digest
        hash_value = hash.hexdigest()

        # checks if both password (hashes) match
        passwords_match = hash_value == base_password_value

    # returns if both password match
    return passwords_match

def password_strength(password):
    """
    Calculates the "theoretical" password strength
    from the given password.
    The returned value is an integer ranging from the lowest
    zero value (unsafest) to a limit value (safest).

    :type password: String
    :param password: The password to be measured for strength.
    :rtype: int
    :return: An integer describing the strength
    level of the given password.
    """

    # starts the strength value
    # counter to the minimum value (zero)
    strength_value = 0

    # retrieves the length of the password
    password_length = len(password)

    # in case the password is not set
    # (empty password)
    if password_length < 1:
        # returns the strength value
        # immediately
        return strength_value

    # increments the strength value
    strength_value += 1

    # in case the password length is less
    # than a minimum of four
    if password_length < 4:
        # returns the strength value
        # immediately
        return strength_value

    # in case the password length is more
    # or equal to eight
    if password_length >= 8:
        # increments the strength value
        strength_value += 1

    # in case the password length is more
    # or equal to eleven
    if password_length >= 11:
        # increments the strength value
        strength_value += 1

    # in case the password contains at least
    # a number in it
    if NUMBER_REGEX.search(password):
        # increments the strength value
        strength_value += 1

    # in case the password contains both lower case and
    # upper case letters
    if LETTER_LOWER_REGEX.search(password) and LETTER_UPPER_REGEX.search(password):
        # increments the strength value
        strength_value += 1

    # in case the password contains special characters
    # in it (extra security)
    if SPECIAL_CHARACTER_REGEX.search(password):
        # increments the strength value
        strength_value += 1

    # returns the strength value
    return strength_value

def md5_crypt(password, salt, magic = DEFAULT_MD5_CRYPT_MAGIC):
    """
    Runs the MD5 crypt algorithm for the given password,
    salt and magic value.
    The magic value is set by default and should be changed
    carefully.

    :type password: String
    :param password: The password to be encrypted.
    :type salt: String
    :param salt: The salt to be used in encryption.
    :type magic: String
    :param magic: The magic value to be used in encryption.
    :rtype: String
    :return: The resulting MD5 crypt value.
    """

    # creates the first hash value
    hash = hashlib.md5()

    # appends the password with the magic value and the salt
    # creating the "appended" value
    appended_value = password + magic + salt

    # updates the hash with the appended value
    hash.update(appended_value)

    # appends the password with the salt and the magic value
    # creating the "new appended" value
    appended_value = password + salt + password

    # retrieves the mixin hash from the appended value
    mixin_hash = hashlib.md5(appended_value)

    # retrieves the mixin digest from the
    # mixin hash
    mixin_digest = mixin_hash.digest()

    # retrieves the password length
    password_length = len(password)

    # iterates over the range of the password
    # length
    for index in range(password_length):
        hash.update(mixin_digest[index % 16])

    # retrieves the password length
    password_length = len(password)

    # retrieves the password (first) character
    password_character = password[0]

    # iterates while the password
    # length is still valid
    while password_length:
        # in case the password length (index)
        # is of type odd
        if password_length & 1:
            # updates the hash with the null value
            hash.update("\0")
        # otherwise the password length (index)
        # is of type even
        else:
            # updates the hash with the
            # password (first) character
            hash.update(password_character)

        # shifts the password length one
        # bit to the right
        password_length >>= 1

    # retrieves the has digest
    hash_digest = hash.digest()

    for index in range(1000):
        # creates the extras hash
        extra_hash = hashlib.md5()

        # in case the index is odd
        if index & 1:
            extra_hash.update(password)
        # in case the index is even
        else:
            extra_hash.update(hash_digest)

        # checks index for modulus three
        if index % 3:
            extra_hash.update(salt)

        # checks index for modulus seven
        if index % 7:
            extra_hash.update(password)

        # in case the index is odd
        if index & 1:
            extra_hash.update(hash_digest)
        # otherwise it must be even
        else:
            extra_hash.update(password)

        # retrieves the hash digest from
        # the extra hash
        hash_digest = extra_hash.digest()

    # creates the rearranged buffer
    rearranged_buffer = []

    # retrieves the various values from a pre-defined set of position
    for a, b, c in ((0, 6, 12), (1, 7, 13), (2, 8, 14), (3, 9, 15), (4, 10, 5)):
        value = ord(hash_digest[a]) << 16 | ord(hash_digest[b]) << 8 | ord(hash_digest[c])

        # iterates over the range of four
        for _index in range(4):
            # converts the value to ascii
            value_ascii = INTEGER_TO_ASCII_64[value & 0x3f]

            # adds the ascii value to the rearranged buffer
            rearranged_buffer.append(value_ascii)

            # shifts the value six bits to the right
            value >>= 6

    # retrieves the "last" value
    value = ord(hash_digest[11])

    # iterates over a two size range
    for _index in range(2):
        # converts the value to ascii
        value_ascii = INTEGER_TO_ASCII_64[value & 0x3f]

        # adds the ascii value to the rearranged buffer
        rearranged_buffer.append(value_ascii)

        # shifts the value six bits to the right
        value >>= 6

    # retrieves the rearranged from by joining
    # rearranged buffer
    rearranged = "".join(rearranged_buffer)

    # creates the MD5 crypt value appending
    # the magic with the salt, the MD5 crypt separator
    # and the rearranged value
    md5_crypt_value = magic + salt + MD5_CRPYT_SEPARATOR + rearranged

    # returns the MD5 crypt value
    return md5_crypt_value

def generate_hash_digest_map(file_path, hash_set = DEFAULT_HASH_SET):
    """
    Generates a map containing a set of hash digests generate
    from the file contained in the given file path.
    The set of hash function to be used may be controlled using the
    hash set parameter.

    :type file_path: String
    :param file_path: The path to the file to be used for hash
    digest calculation.
    :type hash_set: Tuple
    :param hash_set: The set of hash functions to be used.
    :rtype: Dictionary
    :return: The map containing the hash digest values for the file.
    """

    # creates the list to hold the various
    # hash objects
    hash_list = []

    # iterates over the hash set
    # to create the various hash objects
    for hash_name in hash_set:
        # creates the hash object and adds
        # it to the list of hash objects
        hash = hashlib.new(hash_name)
        hash_list.append(hash)

    # opens the file for read
    file = open(file_path, "rb")

    try:
        # iterates continuously
        while True:
            # reads "some" file contents
            file_contents = file.read(4096)

            # in case the file contents are
            # not valid (end of file)
            if not file_contents:
                # breaks the loop
                break

            # iterates over all the hash objects
            # in the hash list to update them
            for hash in hash_list:
                # updates the hash object
                # with the file contents
                hash.update(file_contents)
    finally:
        # closes the file
        file.close()

    # creates the map to hold the various
    # hash digests
    hash_digest_map = {}

    # iterates over all the hash objects
    # in the hash list to retrieve the digest
    # and update the hash digest map
    for hash in hash_list:
        # retrieves the name of the hash (function)
        hash_name = hash.name

        # retrieves the hash (hexadecimal) digest
        hash_digest = hash.hexdigest()

        # updates the hash digest map with the
        # new hash digest
        hash_digest_map[hash_name] = hash_digest

    # returns the hash digest map
    return hash_digest_map
